
from pyspec.client import SpecConnection
from pyspec.client.SpecCommand import SpecCommandA
import time

sp = SpecConnection('toto')
cmd = SpecCommandA(sp,'testcnt')

def titi(value):
    cmd(value)



#
#  @(#)__init__.py	6.3  12/13/20 CSS
#  "pyspec" Release 6
#

import sys
import os

_pyspec_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(_pyspec_dir)

from VERSION import getVersion, getFullVersion

__version__ = getVersion()
__fullversion__ = getFullVersion()





from pyspec.css_logger import log
import numpy as np

try:
    import h5py
    h5_imported = True
except ImportError:
    log.log(2,"hdf5 needs h5py python module install in the system")
    h5_imported = False

import os
import re

class H5Reader(object):
    def __init__(self, filename):
        self.filename = filename

    def get_nb_images(self):
        with h5py.File(self.filename) as ifd:
            return ifd['/entry/data/'].attrs['nb_images']

    def get_image(self, imgno):
        with h5py.File(self.filename) as ifd:
            nb_images = ifd['/entry/data/'].attrs['nb_images']
            if imgno > 0 and imgno <= nb_images:
                dgroup = ifd['/entry/data']
                for ky in dgroup.keys():
                    d = dgroup[ky]
                    fimg = d.attrs['image_nr_low']
                    limg = d.attrs['image_nr_high']
                    if imgno >= fimg and imgno <= limg:
                        idx = imgno - fimg
                        return d[()][idx]

        return None

class H5Writer(object):
    def __init__(self, filename, images_per_file=5, overwrite=False):
        self.filename = filename
        self.overwrite = overwrite
        self.nr_imgs_per_file = images_per_file
        if not h5_imported:
            log.log(2,"Cannot import h5py module. hdf5 disabled")
            self.cansave = False
        else:
            self.cansave = True

    def create(self):
        if not self.cansave:
            return

        if not self.overwrite: 
            self.filename = choose_filename(self.filename)
    
        log.log(2,"Creating file: %s" % self.filename)
        
        with h5py.File(self.filename,'w') as ofd:
            ofd.create_group('/entry/data')
     
    def save_metadata(self, metadata, rootky='/entry'):
        if not self.cansave:
            return
        self.metadata = metadata
        with h5py.File(self.filename,'r+') as fd:
            self._save_metadata(fd, metadata, rootky)

    def _save_metadata(self, fd, values, rootky='/entry'):
        for ky, value in values.items():
            h5ky = os.path.join(rootky,ky)

            if isinstance(value, dict):
                try:
                    fd.create_group(h5ky)
                except:
                    pass
                self._save_metadata(fd, value, h5ky)
            else:
                print("looking for %s in %s" % (ky,rootky))
                if ky in fd[rootky].keys():
                    print("%s already exists" % ky)
                    if isinstance(value, str):
                        del fd[h5ky]
                        fd.create_dataset(h5ky,data=value)
                    else:
                        fd[h5ky].write_direct(np.array([value]))
                else:
                    print("%s creating" % ky)
                    if isinstance(value, str):
                        dtype = 'S'
                    else: # valid for single number values and for list/tuple objects
                        value = np.array(value)
                        dtype = value.dtype
                    fd.create_dataset(h5ky,data=value)

    def save_img(self, data):
        if not self.cansave:
            return

        total_images = 0
        dkeys = []

        fileno = 0
        with h5py.File(self.filename,'r+') as ofd:
            data_group = ofd['/entry/data']

            for ky in data_group.keys() :
                fileno += 1
                dkeys.append(ky)
                last_imgno = data_group[ky].attrs['image_nr_high']

            # how many images in last one
            newfile = False
            if dkeys:
                dataky = os.path.join('/entry/data',dkeys[-1])
                odata = ofd[dataky][()]
                nb_images = len(odata)
                if nb_images >= self.nr_imgs_per_file: 
                    newfile = True
                    fileno += 1
                imgno = last_imgno + 1
            else:
                newfile = True
                fileno = 1
                imgno = 1

            dataky = 'data_%06d' % fileno
            fullky = os.path.join('/entry/data', dataky)

            if newfile:
                data_filename = self.create_datafile(dataky,data=data)
                ofd[fullky] = h5py.ExternalLink(os.path.basename(data_filename), \
                        '/entry/data/data')
                ofd[fullky].attrs.create('image_nr_low', imgno)
                ofd[fullky].attrs.create('image_nr_high', imgno)
            else: # append data to latest file
                dset = ofd[fullky]
                newlen = dset.shape[0]+1
                newshape = (newlen,dset.shape[1], dset.shape[2])
                dset.resize(newshape)
                dset[newlen-1] = data
                ofd[fullky].attrs.modify('image_nr_high',imgno)
            ofd['/entry/data'].attrs.modify('nb_images', imgno)

    def create_datafile(self, dataky, data):
        data_filename = self.filename.replace('master',dataky)
        log.log(2, " - creating datafile %s" % data_filename)
        with h5py.File(data_filename, 'w') as dfd:
            maxshape = tuple((None,)+data.shape)
            shape = tuple((1,)+data.shape)
            dfd.create_dataset('/entry/data/data', data=data, shape=shape, dtype=data.dtype, maxshape=maxshape)
        return data_filename

def choose_filename(filename):
    if not os.path.exists(filename):
        return filename

    # search for _master
    mat=re.search("(.*?)(([\-\.\_]master)(.*?)$)",filename)
    if mat:
        base = mat.group(1)
        master = mat.group(2)
    else:
        base,master= os.path.splitext(filename)
        master = "." + master
    
    fileno = 1
    while os.path.exists(filename):
        filename = "%s_%02d%s" % (base,fileno,master)
        fileno += 1

    return filename

def test():
    import time
    import numpy as np

    metadata = {
            'instrument': {
                    'chi': 23,
                    'phi': 12,
                    'th': 100,
                    'scanpositions': [7,2,3,4],
                },
            'save_time': time.time(),
            'detector': {
                'exptime': 1.1,
                'threshold':  111,
                'acqmode': 'PC',
                'name': 'StreakCamera',
               },
            }

    data = np.zeros((100,100), dtype='f')

    w = H5Writer('/tmp/multi_master.h5')
    w.create()
    w.save_metadata(metadata)
    for i in range(23):
        d = data + i
        w.save_img(d)

if __name__ == '__main__':
    log.start()
    test()



#!/usr/bin/env python3
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%

#%.src: rst/%.rst
#        rst2html.py --template=page.tpl $(RSTFLAGS) $< $@
#        ./postproc.py $@


"""
A minimal front end to the Docutils Publisher, producing HTML.
"""
TPL_HELPPAGE = "spec_help.tpl"

import sys
import re
import os

try:
   from docutils.core      import publish_cmdline, publish_file, Publisher
   from docutils.writers   import html4css1
   from docutils import nodes     
   from docutils.nodes     import reprunicode
   from docutils.writers.html4css1 import HTMLTranslator
except ImportError:
   print("Cannot import docutils")
   sys.exit(1)

if "-check" in sys.argv:
   sys.exit(0)

related_block = """
 <div class="related">
     <h3>%s</h3>
     <ul>
       %s
     </ul>
 </div>
"""

related_index_link = '      <li><a href="index.html">help index</a></li>\n'
related_link = '      <li><a href="%(url)s">%(name)s</a></li>\n'

content_link = '<li><a href="#%s">%s</a></li>\n'

class SpecHTMLTranslator(HTMLTranslator):

   def __init__(self,document):

       HTMLTranslator.__init__(self,document)

       self.in_related    = 0
       self.related       = ""
       self.related_links = ""

       #self.restro = re.compile(r"\*\*(?P<emph>[*()#+/a-zA-Z0-9\[\]\-_]+)\*\*")
       #self.remph  = re.compile(r"\*(?P<emph>[()#+/a-zA-Z0-9\-_]+)\*")

       # (?<!\\)*  means asterisk not preceded by a \
       self.restro = re.compile(r"\*\*(?!\ )(?P<emph>[\\@=\-()\:#+/a-zA-Z0-9_\ ]+)(?<!\ )\*\*")
       self.remph  = re.compile(r"\*(?!\ )(?P<emph>[\\@=\-()\:#+/a-zA-Z0-9_\ ]+)(?<!\ )\*")

       self.contentsbody = ""

   def visit_document(self,node):
       HTMLTranslator.visit_document(self,node)
       self.doctitle = node.get('title')

   def visit_subtitle(self,node):
       HTMLTranslator.visit_subtitle(self,node)
       self.docsubtitle = node.get('names')[0]

   def visit_section(self,node):

       try:
         nodename = reprunicode(node['names'][-1]).lower()

         try:
               idanchor = node.__dict__['attributes']['ids'][0]
               contentlink = content_link % (idanchor, nodename )
               self.contentsbody += contentlink 
         except:
               print("cannot get content info for node "+nodename)

         if nodename.find("related")  == 0 or \
            nodename.find("other")    == 0 :
               node['classes'].append('related-section')
               self.in_related = 1
               self.related_title = nodename
         elif nodename.find("see also") == 0:
               node['classes'].append('see-also')
               self.in_related = 1
               self.related_title = nodename
       except:
          pass 

       HTMLTranslator.visit_section(self,node)

   def depart_section(self,node):
      if self.in_related == 1:
         if self.related_title == "see also":
            self.related_links += related_index_link 
         self.related += related_block % (self.related_title.capitalize(), self.related_links)
         self.related_links = ""
         self.related_title = ""
         self.in_related = 0

      HTMLTranslator.depart_section(self,node)

   def depart_table(self, node):
       if len(self.context):
           self.compact_p = self.context.pop()
       self.body.append("</table>\n")

   def visit_reference(self,node):
       HTMLTranslator.visit_reference(self,node)
       if self.in_related:
          if 'name' in node.__dict__['attributes'].keys():
              vals = {'url': node.get('refuri'), 'name': node['name'] }
          else: 
              vals = {'url': node.get('refuri'), 'name': node.get('refuri') }
          self.related_links += related_link % vals


   def depart_document(self,node):
      if self.related_links != "":
         self.related += related_block % self.related_links

      self.contents = "<ul>\n%s</ul>\n" % self.contentsbody

      HTMLTranslator.depart_document(self,node)


   """
    Tables without borders
    - amazingly the border is set to 1 in the html4css1 writer, and not configurable 
      through css
   """
   def visit_table(self, node):
      classes = ' '.join(['docutils', self.settings.table_style]).strip()
      self.body.append(
           self.starttag(node, 'table', CLASS=classes, border="0"))

   """
    Rewrite the whole visit_literal from the original docutils 0.8.1
    html4css1 writer. 
    
    This solves two issues:
    - stop using tt that is a not supported tag in HTML-5 (span is used instead)
    - allow for emphasis inside literal blocks (necessary for spec and c-plot
      function parameters.  This functionality should actually be part of the parser
      but it would affect too many classes and compromise later compatibility
   """ 
   def visit_literal(self, node):
        """
        Inside literal will replace a string of letter, numbers and hyphens surrounded by * to be emphasized
        """ 

        """Process text to prevent tokens from wrapping."""
        self.body.append(
            self.starttag(node, 'span', '', CLASS='docutils literal'))
        text = node.astext()
        for token in self.words_and_spaces.findall(text):
            if token.strip():
                # Protect text like "--an-option" and the regular expression
                # ``[+]?(\d+(\.\d*)?|\.\d+)`` from bad line wrapping
		#if self.sollbruchstelle.search(token):
		#    toappend = '<span class="pre">%s</span>' % self.encode(token)
		#else:
                toappend = self.encode(token)
            elif token in ('\n', ' '):
                # Allow breaks at whitespace:
                toappend = token
            else:
                # Protect runs of multiple spaces; the last space can wrap:
                toappend = '&nbsp;' * (len(token) - 1) + ' '
            
            toappend = self.process_Text(toappend)
            self.body.append(toappend)
        self.body.append('</span>')
        # Content already processed:
        raise nodes.SkipNode

   def visit_literal_block(self, node):

        self.body.append(self.starttag(node, 'pre', CLASS='literal-block'))

        text = node.astext()
        text = self.process_Text(text)

        self.body.append( text )

        self.body.append('\n</pre>\n')

        # Content already processed:
        raise nodes.SkipNode

   def depart_literal_block(self, node):
        pass

   def process_Text(self, text):
       retstr = text

       # reserve escaped backlashes (in text as \\)
       retstr = re.sub("\\\\\\\\","&7&", retstr)

       # reserve escaped asterisks (in text as \*)
       retstr = re.sub("\\\\\*", "@=@", retstr)

       retstr = self.restro.sub("<b>\g<emph></b>",retstr )
       retstr = self.remph.sub("<i>\g<emph></i>",retstr )

       # restore escaped asterisks as normal asterisks
       retstr = re.sub("@=@", "*", retstr)

       # restore escaped backslashes as normal backslashes
       retstr = re.sub("&7&", "\\\\",  retstr)

       return retstr

class SpecHTMLWriter(html4css1.Writer):

    def __init__(self, template=None):
      html4css1.Writer.__init__(self)
      self.translator_class = SpecHTMLTranslator
      self.template_str = ""
      self.extraclass   = None

      if template:
          self.setTemplate(template)

    def setTemplate(self, templ_file):
      self.template_file = templ_file
      if os.path.exists(self.template_file):
          self.template_str = open(self.template_file).read()
      else:
          self.template_str = ""

    def process(self, filename, outfilename="/dev/null", template=None, extraclass=None):

       if template:
           self.setTemplate(template)

       self.extraclass = extraclass
       convbuf = publish_file( source_path=filename, destination_path=outfilename, writer_name='html',  writer=self) 
       return(self.doctitle, self.docsubtitle, convbuf)

    def apply_template(self):

      if not self.template_str:
         print("NO TEMPLATE")
         raise BaseException("NoTemplate")

      subs = self.interpolation_dict()

      subs['related']     = self.visitor.related
      subs['description'] = self.visitor.doctitle
      subs['contents']    = self.visitor.contents 

      return self.template_str % subs

    def translate(self):
        html4css1.Writer.translate(self)
        self.doctitle    = self.visitor.doctitle
      
        try:
            self.docsubtitle = self.visitor.docsubtitle
        except:
            self.docsubtitle = "NO SUBTITLE"

def dOneFile(filename, template=None):

    from os.path import dirname, abspath

    if template is None:
        template = os.path.join(dirname(abspath(__file__)), "spec_help.tpl")

    writer = SpecHTMLWriter(template=template)
    return writer.process( filename )

def process(filename, outfile=None, template=None):
    buf = open(filename).read()
    title, subtitle, converted = dOneFile(filename , template)

    if not outfile:
       fd = sys.stdout
    else:
       fd = open(outfile,"w")

    fd.write(converted)

if __name__ == '__main__':

    import sys

    if "-d" in sys.argv:
        idx = sys.argv.index("-d")
        sys.argv.pop(idx)
        DEBUG = 1

    if len(sys.argv) < 2:
        print("Usage: %s filename [outfile]", sys.argv[0])
        sys.exit(0)

    filename = sys.argv[1]

    if len(sys.argv) >= 3:
        outfile = sys.argv[2]
    else:
        outfile = None

    process(filename, outfile)




#******************************************************************************
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%
#
#  Copyright (c) 2013,2014,2015,2016,2017,2019,2020,2021
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("pyspec") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
#******************************************************************************

import os
from graphics_rc import g_rc

try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QDialog, QVBoxLayout, QGridLayout, \
        QLabel, QComboBox, QHBoxLayout, QRadioButton, \
        QPushButton, QApplication, QColorDialog, QMainWindow, \
        QWidget, QButtonGroup, QSplashScreen, \
        QListWidgetItem, QListWidget, QTreeWidgetItem, QTreeWidget, \
        QSplitter,  QTabWidget, QMenuBar, QMessageBox, QSpacerItem,  \
        QLineEdit, QHeaderView, QSpinBox, QTextEdit, QTextBrowser,  \
        QMenu, QAction, QTabBar, QStackedWidget,  QFileDialog,  \
        QScrollArea, QProgressBar, QStyleOptionSlider, QGroupBox,  \
        QAbstractItemView, QSizePolicy, QFrame, QStyle, QScrollBar, QToolBar, \
        QDialogButtonBox, QToolButton, QCheckBox, QLayout, QInputDialog

    from PyQt5.QtGui import QFont, QIcon, QFontMetrics, QPen, QColor, QPainter, \
        QPixmap, QBrush, QPainterPath , QDesktopServices 

    from PyQt5.QtGui import QIntValidator, QRegExpValidator, QDoubleValidator, \
            QValidator

    from PyQt5.QtCore import QObject, QTimer,  QProcess, QEvent, QSize, \
         QLine, QRect, QPoint, QRectF, QUrl, QUrlQuery, QCoreApplication, \
         pyqtSignal

    from PyQt5.QtPrintSupport import QPrinter, QPrintDialog
    from PyQt5.QtCore import QT_VERSION_STR

    getQApp = QCoreApplication.instance

    g_rc.qt_imported = True
    g_rc.qt_variant = 'PyQt5'
    g_rc.qt_version = list(map(int, QT_VERSION_STR.split(".")))
except:
    pass




#
#  @(#)graphics_rc.py	6.1  05/11/20 CSS
#  "pyspec" Release 6
#

import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

class g_rc(object):
    qt_variant = None
    qt_version = None
    graph_variant = None

    qt_imported = False
    mpl_available = False
    mpl_imported = False
    qwt_imported = False

    mpl_version = None
    mpl_version_no = None



#******************************************************************************
#
#  @(#)spec_shm.py	6.1  11/02/20 CSS
#
#  "splot" Release 6
#
#  Copyright (c) 2020
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("splot") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
#******************************************************************************

import time

from pyspec.css_logger import log

from pyspec import datashm

# provide through this module all functionalities of datashm
from pyspec.datashm import getdata, getspeclist
from pyspec.datashm import getinfo, getdatacol, getdatarow
from pyspec.datashm import getarrayinfo, getarraylist
from pyspec.datashm import getmetadata

class _UpdateTable(object):

    def __init__(self):
        self.created = time.time()
        self._update = {}

    def isupdated(self, spec, var, client):
        if client not in self._update:
            self._update[client] = {}
        if spec not in self._update[client]:
            self._update[client][spec] = {}
        if var not in self._update[client][spec]:
            self._update[client][spec][var] = True

        if datashm.isupdated(spec, var):
            for clnt in self._update:
                try:
                    self._update[clnt][spec][var] = True
                except:
                    pass

        retval = self._update[client][spec][var]
        self._update[client][spec][var] = False
        return retval

class UpdateTable(_UpdateTable):

    def __new__(cls):
        if not hasattr(cls, '_inst'):
            cls._inst = super(UpdateTable, cls).__new__(cls)
        else:
            def init_pass(self, *dt, **mp): pass
            cls.__init__ = init_pass

        return cls._inst

global utbl
utbl = UpdateTable()

def is_updated(spec, var, client):
    return utbl.isupdated(spec, var, client)




from pyspec.client.SpecConnection import SpecConnection, QSpecConnection
from pyspec.css_logger import log
from pyspec.client.spec_updater import spec_updater
from pyspec.utils import async_loop
import time

from pyspec.client import spec
spec = spec("toto")

updater = spec_updater()
updater.start()

while True:
   spec.run_cmd("testcnt 1;")
   time.sleep(0.1)

"""
global s

def run_mac():
    global s
    answ = s.run_cmd("graph toto;")
    log.log(2, "got answer %s" % answ)

def update_conn():
    global s
    async_loop() 
    if not s.is_connected():
        log.log(2, "not connected")
    s.update_events()

s = QSpecConnection("toto")

time.sleep(2)

while True:
   run_mac()
   time.sleep(1)



"""



#******************************************************************************
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%
#
#  Copyright (c) 2017,2020,2021
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("pyspec") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
#******************************************************************************

import sys
import platform
import os
import socket
import asyncore

def is_macos():
    return sys.platform == "darwin" 

def is_windows():
    return sys.platform == "win32" 

def is_unity():
    desktop_session = os.environ.get("DESKTOP_SESSION", None) 
    if desktop_session in ["ubuntu-2d", "ubuntu"]:
        return True
    else:
        return False

def is_remote_host(host):
    if host == 'localhost' or host is None:
         return False

    local_ip = socket.gethostbyname(socket.gethostname())
    host_ip = socket.gethostbyname(host)

    if local_ip == host_ip:
        return False
    else:
        return True

def is_python2():
    return sys.version_info[0]== 2

def is_python3():
    return sys.version_info[0] == 3

def is_centos8():
    import platform
    linux_dist = platform.linux_distribution()
    if linux_dist[0].lower().find('centos') != -1 and \
        linux_dist[1][0] == '8':
            return True
    else:
        return False

def async_loop(timeout=0.01, use_poll=True, count=None):
    """Start asyncore and scheduler loop.
    Use this as replacement of the original asyncore.loop() function.
    """
    if use_poll and hasattr(asyncore.select, 'poll'):
        poll_sock = asyncore.poll2
    else:
        poll_sock = asyncore.poll

    sockmap = asyncore.socket_map

    if count is None:
        while sockmap:
            poll_sock(timeout, sockmap)
    else:
        while sockmap and count > 0:
            poll_sock(timeout, sockmap)
            count -= 1

if __name__ == '__main__':
   print("MacOS: ", is_macos())
   print("Ubuntu Unity: ", is_unity())
   print("Windows: ", is_windows())
   print("Python 2: ", is_python2())
   print("Python 3: ", is_python3())




#!/usr/bin/env python

#   %W%  %G% CSS
#
#   "spec" Release %R%
#
#   Copyright (c) 2018,2019,2020,2021,2022
#   by Certified Scientific Software.
#   All rights reserved.
#   Copyrighted as an unpublished work.

import asyncore
import socket
import time
import sys
import os
import optparse
import re
import struct
import signal
import logging

from logging.handlers import RotatingFileHandler

progname = os.path.basename(sys.argv[0])

class HardwareLogger(logging.Logger):
    def isEnabledFor(self, level):
        return level <= self.getEffectiveLevel()

log = HardwareLogger(__name__)
log.setLevel(2) # by default log messages with level <= 2

__version__ = "1.0"
protocol_version = "V2"
LOCK_DIR = '/tmp'
LOCK_FILE = ''

re_cmd = re.compile("=: (?P<cmdno>\d+) (?P<cmd>.+?)$")

STATUS_BUSY, STATUS_IDLE = (1,0)

struct_types = {
    "ubyte": {'type': "B", 'size': 2},
    "ushort": {'type': "H", 'size': 2},
    "ulong": {'type': "L", 'size': 4},
    "ulong64": {'type': "Q", 'size': 8},
    "byte": {'type': "b", 'size': 2},
    "short": {'type': "h", 'size': 2},
    "long": {'type': "l", 'size': 4},
    "long64": {'type': "q", 'size': 8},
    "float": {'type': "f", 'size': 4},
    "double": {'type': "d", 'size': 8},
}

def wrap_up(*args):
    if os.path.exists(LOCK_FILE):
        os.remove(LOCK_FILE)
    sys.exit(0)

class Command(object):
    def __init__(self, cmdno, cmd):
        self.cmdno = cmdno
        self.cmd = cmd
        self.response = None
        self.address = 0
        self.error = ''

        parts = cmd.split()

        self.cmdname = parts[0]
        if len(parts) > 1:
            self.args = parts[1:]
        else:
            self.args = []

        argno = 0
        address = None
        for arg in self.args:
            if arg.startswith("a="):
                address = arg[2:]
                break
            argno += 1

        if address:
            self.args.pop(argno)
            self.address = address

    def get_error(self):
        return self.error

    def set_error(self, error):
        self.error = error

    def get_command_name(self):
        return self.cmdname

    def get_address(self):
        return self.address

    def get_args(self):
        return self.args

    def set_response(self, response):
        self.response = str(response)

    def get_command_no(self):
        return self.cmdno

    def get_response(self):
        return self.response

    def get_response_length(self):
        if self.response:
            return len(self.response)
        else:
            return 0

class CommandHandler(asyncore.dispatcher):

    def __init__(self, *args):
        asyncore.dispatcher.__init__(self,*args)
        self.pending_commands = []
        self.writebuf = ''
        self.databuf = ''
        self.readbuf = ''
        self.last_cmdno = -1
        self.data_expected = 0
        self.received_data = ''
        self.sending_data = False
    
    def command_waiting(self):
        return True if self.pending_commands else False

    def get_command(self):
        if self.pending_commands:
            cmd = self.pending_commands.pop(0)
            return cmd

        return None

    def send_response(self,cmd):
        error = cmd.get_error()
        cmdno = cmd.get_command_no()
        resp = cmd.get_response()
        rlen = cmd.get_response_length()

        if error:
            first_char = "!"
        else:
            first_char = "@"

        if rlen: 
            outstr = "%s: %d %d#%s\n" % (first_char, cmdno, rlen, resp)
        else:
            outstr = "%s: %d %d#\n" % (first_char, cmdno, rlen)
        self.writebuf += outstr

    def send_data(self,strdata):
        log.log(4,"sending data") 
        log.log(4," strdata is %s" % type(strdata))
        self.sending_data = True
        self.databuf = strdata

    def readable(self):
        return True

    def writable(self):
        if self.writebuf or self.sending_data:
            return True
        else:
            return False

    def handle_read(self):
        newdata = self.recv(32768)

        # check first if it should be data
        if self.data_expected:
            lendata = len(newdata)
            if lendata <= self.data_expected:
                buf = newdata
                cmd_buf = ''
            else:
                buf = newdata[:self.data_expected]
                cmd_buf = newdata[self.data_expected:]

            self.data_expected -= len(buf)
            self.received_data = ''.join([self.received_data, buf])

            got = len(self.received_data)
            log.log(3, "receiving data.  Got %d so far. %d still missing", got, self.data_expected)

            if self.data_expected != 0:
                log.log(3, "keep waiting for data")
                return
        else:
            cmd_buf = newdata

        if sys.version_info[0] >= 3:
            self.readbuf += cmd_buf.decode('utf-8')
        else:
            self.readbuf += cmd_buf

        # separator is newline. several commands or only part of it can arrive in one read
        while True: 
            eoc = self.readbuf.find('\n')    

            if eoc == -1:
                break
            
            command = self.readbuf[:eoc]
            self.readbuf = self.readbuf[eoc+1:]
            # log.log(1, "new command received: %s", command[0:10])

            # is it a valid command?
            valid = self.check_cmd_syntax(command)
            if not valid and not self.data_expected: 
                log.log(1, "Not a valid command: %s", command)
                continue
            else:
                cmdno, cmd = valid
                if cmdno == self.last_cmdno:
                    log.log(3, "cmdno: %s - already handled. Ignored", cmdno)
                    continue
                else:
                    log.log(4, "cmdno: %s - new command:  [\"%s\"]", cmdno, cmd)
                    self.pending_commands.append( Command(cmdno, cmd) )
                    self.last_cmdno = cmdno

    def handle_write(self):

        if self.writebuf:
            if sys.version_info[0] >= 3:
                sent = self.send( self.writebuf.encode() )
            else:
                sent = self.send( self.writebuf )

            self.writebuf = self.writebuf[sent:]
            log.log(4, "%s values sent. %s remain.", sent,len(self.writebuf))

        if self.sending_data: # data is always sent raw - as bytes
            log.log(4, "sending data")
            sent = self.send( self.databuf )
            self.databuf = self.databuf[sent:]
            if self.databuf:
                log.log(4, "%s values sent. %s remain.", sent,len(self.databuf))
            else:
                log.log(4,"all data sent")
                self.sending_data = False

    def handle_close(self):
        log.log(2, "Connection closed")
        self.close()

    def check_cmd_syntax(self,cmdbuf):
        mat = re_cmd.match(cmdbuf)
        if mat:
            cmdno = int(mat.group('cmdno'))
            cmd = mat.group('cmd')
            return cmdno, cmd

        return None

    def get_incoming_data(self):
        return self.received_data

    def expect_data(self, data_size):
        self.received_data = ''
        self.data_expected = data_size
        log.log(1, "expecting data  size: %d", self.data_expected)

class HardwareServer(asyncore.dispatcher):

    doc = """
Base class for hardware servers.

Check documentation for writing your own server
"""

    default_port = 5000
    description = "generic"
    dtype = None

    def __init__(self):
        global LOCK_FILE

        self.opt_parser = None
        self.hdw_pars = {}
        self.i_am_running = True
        self.say_bye = False
        self.data_pending = False  # whether there is data to be sent to client
        self.data_acquired = False # whether any data has been acquired
        self.data_read = False # whether data has been read from hardware
        self.data_buffer = None
        self.pending_incoming_data = False # whether expecting data from a write command

        self.incoming_roi = False
        self.roi_table = {}
        self.dims = 1

        self.server_commands = {
            'hello':  self.say_hello,
            'config':  self.config_hdw,
            'exit':  self._server_exit,
            'clear':  self.server_clear,
            'run':  self.acq_run,
            'halt':  self.acq_halt,
            'get_status':  self.get_status,
            'set':  self.set_par,
            'get':  self.get_par,
            'read':  self.read_data,
            'write':  self.write_data,
            'goodbye':  self.goodbye,
            'xfer_done':  self.xfer_done,
            'set_roi': self.set_roi,
            'get_roi': self.get_roi,
        }

        self.configure_server()
        self.parse_args()
        self.handle_args()

        self.selected_address = None

        self.port = self.settings.port
        self.logfile = self.settings.logfile
        LOCK_FILE = self.settings.lockfile

        if self.settings.do_daemon:
            if not self.logfile:
                self.logfile = "/tmp/%s-%s.log" % (progname,self.port)

        # setup logging 
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d - %(levelno)s - %(message)s', \
                         datefmt='%H:%M:%S')

        if self.settings.verbose_level:
            log.setLevel(self.settings.verbose_level)

        if self.logfile:
            #logh = logging.FileHandler(self.logfile)
            logh = RotatingFileHandler(self.logfile,maxBytes=20000000,backupCount=5)
            log.addHandler(logh)
            logh.setFormatter(formatter)
            log.log(3, "Logging begun on logfile %s", self.logfile)
        elif not self.settings.do_kill:
            logh = logging.StreamHandler()
            log.addHandler(logh)
            logh.setFormatter(formatter)
            log.log(3, "Logging begun on stdout")
        # setup logging end

        if not LOCK_FILE:
            LOCK_FILE = "%s/%s-%s.lock" % (LOCK_DIR, progname,  self.port)

        ret = self.check_lock()
        if ret:
            if ret == 1:
                log.log(1, "%s is already running", progname)
                wrap_up()

        if self.settings.do_kill:
            wrap_up()

        if self.settings.do_daemon:
            self.daemonize()

        signal.signal(signal.SIGTERM, self._server_exit)

        try:
            if os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()):
                #  Running in foreground
                signal.signal(signal.SIGINT, wrap_up)
            #  dont catch control-C if in background
        except BaseException:
            pass

        self.make_lock()

        self.init_server()

        self.handler_class = CommandHandler
        self.handler = None

        self.cmd = None
        self.status = STATUS_IDLE
        self.executing = False

        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind(('', self.port))
        self.listen(5)

    def configure_server(self):
        # stub. to be defined by implementing classes
        pass

    def init_server(self):
        # stub. to be defined by implementing classes
        pass

    def is_1D(self):
        return self.dims != 2

    def is_2D(self):
        return self.dims == 2

    def set_data_dims(self, dims):
        # function to be called by implementation server
        self.dims = dims

    #  arguments handling
    def parse_args(self):
        parser = self.get_option_parser() 
        self.settings, args = parser.parse_args()

    def handle_args(self):
        # allow implementing classes to act on args 
        pass

    def get_option_parser(self):
        if not self.opt_parser:
            self.set_default_args()
        return self.opt_parser

    def add_option(self, sopt, lopt, **kwargs):
        parser = self.get_option_parser()
        opt = parser.add_option
        opt(sopt,lopt,**kwargs)

    def set_default_args(self):
    
        self.opt_parser = optparse.OptionParser(
            usage="usage: %prog [options]",
            description=self.description,
            version="%%prog %s" % __version__)

        opt = self.opt_parser.add_option
        opt('-p', '--port', dest='port', type='int', default=self.default_port,
	    help = "port number")
        opt('-l', '--logfile', dest='logfile', type='str',
	    help = "if specified, log messages will be saved to logfile")
        opt('-L', '--lockfile', dest='lockfile', type='str', 
	    help = "specifies an alternate lockfile.  If unspecified, lockfile will be placed in /tmp")
        opt('-v', '--verbosity', dest='verbose_level', type='int', default=1,
	    help = "")
        opt('-d', '--demonize', dest='do_daemon', action='store_true', default=False,
	    help = "")
        opt('-k', '--kill', dest='do_kill', action='store_true', default=False,
	    help = "")
    
    def make_lock(self):

        try:
            open(LOCK_FILE, "w").write("%d\n" % self.get_pid())
            return 0
        except BaseException:
            import traceback
            log.log(1, "Cannot create lock file %s", LOCK_FILE)
            log.log(1, traceback.format_exc())
            return -1

    def check_lock(self):
        kill_it = self.settings.do_kill

        if not os.path.exists(LOCK_FILE):
            return 0

        try:
            buf = open(LOCK_FILE, 'r').read().strip()
        except IOError:
            log.log(1, "Cannot read existing lock file %s",  LOCK_FILE)
            return -1

        # read pid from lockfile
        try: 
            pid = int(buf)
        except BaseException:
            log.log(1, "File %s is not a server lock file.\n", LOCK_FILE)
            return -1

        # see if process running
        if not self.check_pid(pid):
            os.remove(LOCK_FILE)
            if kill_it:
                log.log(1, "No existing processs found .\n")
                return 0

        if kill_it:
            os.kill(pid,signal.SIGTERM)
            for _i in range(20):
                if not self.check_pid(pid):
                    break
                time.sleep(0.1)
            else:
                os.kill(pid,signal.SIGKILL)

            try:
                os.remove(LOCK_FILE)
            except BaseException:
                # ignore. probably the killed process has removed it
                pass

            if not self.check_pid(pid): 
                print("Process %d terminated" % pid)
                return 0

            print("Unable to terminate process %d" % pid)
            return -1

        return 1
   
    def check_pid(self,pid):        
        """ Check For the existence of a unix pid. """
        try:
            os.kill(pid, 0)
            time.sleep(0.3)
        except OSError:
            return False
        else:
            return True
           
    def daemonize(self):
        # first fork
        self.fork()

        # detach environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # second fork
        self.fork()

        sys.stdout.flush()
        sys.stderr.flush()

        # redirect stdin, stdout, stderr
        stream = open('/dev/null', 'r')
        os.dup2(stream.fileno(), sys.stdin.fileno())
        stream = open('/dev/null', 'a+')
        os.dup2(stream.fileno(), sys.stdout.fileno())
        stream = open('/dev/null', 'a+')
        os.dup2(stream.fileno(), sys.stderr.fileno())


    def fork(self):
        """
        Spawn the child process
        """
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as exc:
            sys.stderr.write("Fork failed: %d (%s)\n" % (exc.errno, exc.strerror))
            sys.exit(1)

    def get_pid(self):
        return os.getpid()

    def get_hostname(self):
        return socket.gethostname()

    def should_quit(self):
        return not self.i_am_running

    # protocol support
    def handle_cmd(self, cmd):
        cmdname = cmd.get_command_name()
        args = cmd.get_args()
        addr = cmd.get_address()

        log.log(4, "Executing command %s with args %s", cmdname, str(args))

        if cmdname not in self.server_commands:
            log.log(1, "Unsupported command %s", cmdname )
            return("unknown")
        
        if self.executing:
            log.log(1, "Execution in progress.  Command ignored %s", cmdname)
            return("busy")
        
        cmdfunc = self.server_commands[cmdname]

        self.executing = True

        try:
            self.cmd = cmd
            self.select_address(addr)
            retval = cmdfunc(*args) 
            if retval is True:
                retval = 1
            elif retval is False:
                retval = 0
            elif retval is None:
                retval = ""
            return retval
        except BaseException as e:
            errstr = str(e)
            import traceback
            log.log(1, "Error while executing %s. %s", cmdname, errstr)
            log.log(1, traceback.format_exc())
            return "%s. error: %s" % (cmdname, errstr), True

    def handle_accept(self):
        # when we get a client connection start a dispatcher for that
        # client
        sock, addr = self.accept()
        log.log(2, "Connection by %s on port %s", addr, self.port)
        self.handler = self.handler_class(sock)

    def handle_close(self):
        log.log(2, "Connection closed")
        self.close()

    def command_waiting(self):
        if self.handler:
            return True if self.handler.command_waiting() else False

        return False

    def check_command(self):
        if self.command_waiting():
            if self.handler:
                cmd = self.handler.get_command()
                if cmd:
                    return cmd
        return None
      
    def send_response(self, cmd):
        self.handler.send_response(cmd)

    def _watch_cmds(self):
        self.executing = False
        self.set_status(STATUS_IDLE)

    def update(self):
        try:
            if self.executing:
                self._watch_cmds()

            cmd = self.check_command()

            if cmd is not None:
                response = self.handle_cmd(cmd) 
                log.log(4, "cmdno: %s - response is:  %s", cmd.cmdno, str(response))
                if type(response) in (list,tuple):
                    response, err = response
                else:
                    err = False
                cmd.set_response(response)  
                cmd.set_error(err)
                self.send_response(cmd)
                if self.data_pending:
                    self.handler.send_data(self.data_buffer)
                    self.data_pending = False

        except BaseException:
            import traceback
            log.log(1, traceback.format_exc() )

    def run(self):
        self.i_am_running = True 
        self.say_bye = False
    
        while asyncore.socket_map:
            # sync asyncore
            asyncore.loop(timeout=0.1, count=1)
            self.update()
            time.sleep(0.001) # to avoid grabbing 100 cpu

            if self.say_bye:
                log.log(1,"Bye")
                break

            if self.should_quit():
                # let a chance to get the response out
                self.say_bye = True
                continue

        wrap_up()

    # END protocol support
    
    # COMMANDS
    # server public interface
    def set_commands(self, cmds):
        self.server_commands.update(cmds)

    def add_command(self, cmd, fnct):
        self.server_commands[cmd] = fnct

    def set_status(self,status):
        self.status = status

    def set_idle(self):
        self.status = STATUS_IDLE
    def set_busy(self):
        self.status = STATUS_BUSY

    def get_status(self):
        return self.status, False

    def is_idle(self):
        return (self.status == STATUS_IDLE)

    def is_busy(self):
        return (self.status == STATUS_BUSY)

    def get_description(self):
        return self.description, False

    def select_address(self, address):
        self.selected_address = address

    def say_hello(self, name=None):
        fields = {
           'pid': self.get_pid(),
           'host': self.get_hostname(),
           'desc': self.description,
           'pvers': protocol_version,
        }
        
        if name is not None and name != self.my_name:
            if name != self.my_name:
                hello_str = "This isn't the server you are looking for"
                err = True
        else:
            hello_str = "hello back %(pvers)s %(host)s %(pid)s %(desc)s" % fields
            err = False

        return hello_str, err

    def config_hdw(self, address):
        # returns data_type and size
        msg = "Server needs to implement config_hdw. addr: %s"  % address
        log.log(1, msg)
        return msg, True

    def server_exit(self):
        """ to be overriden by implementing classes """
        pass

    def _server_exit(self, *args):
        # quit
        self.server_exit()
        log.log(3, "server_exit called (args=%s)", str(args))
        self.i_am_running = False
        return "ok", False

    def server_clear(self):
        # clear memory
        log.log(3, "server_clear called")
        return "ok", False

    def acq_run(self, preset, mode):
        log.log(3, "acq_run called with preset and mode: %s - %s", str(preset), str(mode))

    def acq_halt(self, mode):
        log.log(3, "acq_halt called with mode: %s", str(mode))

    def set_par(self, *args):
        if args[0].startswith("a="):
            parname, parvalue = args[1:]
        else:
            parname, parvalue = args[0:2]

        self.hdw_pars[parname] = parvalue
        log.log(3, "set_par parname %s saved in memory with value: %s", parname, parvalue)

        return parvalue

    def get_par(self, *args):
        if args[0].startswith("a="):
            parname = args[1]
        else:
            parname = args[0]

        parvalue = self.hdw_pars.get(parname,None)
        log.log(3, "get_par parname %s has value in memory: %s", parname, parvalue)
        return parvalue

    def read_data(self, *args):
        nb_args = len(args)
        log.log(3, "server_read called with args: %s", str(args))

        if nb_args == 2:
            # 1D
            dims = 1
            # first, last = map(int,args)
        elif nb_args == 4:
            # 2D
            dims = 2
            # row_beg, row_end, col_beg, col_end = map(int,args)
        else:
            dims = 0
            log.log(1, "Cannot decode arguments for read_data")
            return "cannot decode args", True

        # get a 1d or 2d numpy array from server

        log.log(3, "Calling get_data")
        data, dtype = self.get_data(*args)

        if dtype is True:  # it is an error. data is not ready
            return "not ready", True

        if dims == 2:
            roi_data = data
            data = roi_data.flatten()
            data = data.tolist()
            log.log(2, "roi data integration in HardwareServer.py is %s", sum(data))
        else: # dims = 1
            pass
            #rows = len(data)
            #roi_data = data[first:last+1]
            #data = roi_data

        self.dtype = dtype
        self.data_buffer = self.pack_data(data)
        self.data_pending = True

        if dims == 2:
            return "%s %s" % ( len(data), self.dtype )
        else: # dims = 1
            nb_points = len(data)
            # nb_points = last - first + 1
            return "%s" % nb_points

    def write_data(self, *args):
        log.log(3,"write_data called with args: %s", str(args))

        dims = map(int, args)
        if self.is_1D():
            nb_data = dims[1] - dims[0] + 1
        else:
            log.log(2,"it is 2D - dims are: %s", str(dims))
            nb_rows = dims[1] - dims[0] + 1
            nb_cols = dims[3] - dims[2] + 1
            nb_data = nb_rows * nb_cols

        self.pending_incoming_data = True
        self.incoming_roi = dims
        len_data_expected = nb_data * 4  # 4 bytes for a long 2 bytes for a chr in a string
        self.handler.expect_data(len_data_expected)
        return "ok"

    def goodbye(self, *args):
        log.log(3, "Client (%s) disconnected", str(args))
        return "ok", False

    def xfer_done(self, *args):
        if self.pending_incoming_data:
            databuf = self.handler.get_incoming_data()
            data_received = self.unpack_data(databuf)
            log.log(3,"data received: %s", str(args))
            self.set_data(data_received,self.incoming_roi)
            self.pending_incoming_data = False

        self.handler.expect_data(0)
        return "ok", False

    # END COMMANDS

    def get_data(self, *args):
        log.log(1, "get_data should be provided by implementation")

    def set_data(self, data, roi_info):
        log.log(1,"set_data should be provided by implementation")

    def set_roi(self, *args):
        log.log(2, "Running set_roi command with args %s", str(args))
        if self.is_2D():
            if len(args) == 5:
                roiname, row_beg, row_end, col_beg, col_end = args
                self.roi_table[roiname] = map(int, [row_beg, row_end, col_beg, col_end])
            else:
                return "Wrong number of values for ROI set", True
        else:
            if len(args) == 3:
                roiname, beg, end = args
                self.roi_table[roiname] = map(int, [beg, end])
            else:
                return "Wrong number of values for ROI set", True

        return ""

    def get_roi(self, *args):
        log.log(2, "Running get_roi command with args %s",str(args))

        if self.data is None:
            return "No acquisition done", True

        if not self.data_read:
            self.get_data()

        if len(args) == 1:
            roiname = args[0]
 
            if roiname not in self.roi_table:
                return "Unknown ROI", True

            if self.is_2D():
                row_beg, row_end, col_beg, col_end = self.roi_table[roiname]
            else:
                roi_beg, roi_end = self.roi_table[roiname]

        elif len(args) == 4 and self.is_2D():
            row_beg, row_end, col_beg, col_end = map(int,args)
        elif len(args) == 2 and self.is_1D():
            roi_beg, roi_end = map(int, args)
        else:
            return "Bad ROI parameters", True

        if self.is_2D():
            return self.data[row_beg:row_end+1,col_beg:col_end+1].sum()

        # 1D
        return self.data[roi_beg:roi_end+1].sum()

    def pack_data(self, data_list):
        dtype = self.type_to_struct(self.dtype)
        fmt = "<%d%s" % (len(data_list), dtype)
        log.log(3,"packing data. %s values. type is %s. fmt: %s", len(data_list), self.dtype, fmt)
        data_str = struct.pack(fmt, *data_list) 

        log.log(4, "length of string is %s", len(data_str))
        return data_str

    def unpack_data(self, data_str):

        dtype = self.type_to_struct(self.dtype)
        dsize = self.type_to_size(self.dtype)
        tsize = len(data_str)/dsize
        log.log(3,"number of things to unpack %s", tsize)

        fmt = "<%d%s" % (tsize, dtype)

        log.log(3,"unpacking data from str length %s. type is %s. fmt: %s", len(data_str), self.dtype, fmt)
        data = struct.unpack(fmt, data_str)
        log.log(3,"unpacked data. got %s values", len(data))

        return data

    def type_to_struct(self,dtype):
        if dtype in struct_types:  
            return struct_types[dtype]['type']
        return None

    def type_to_size(self,dtype):
        if dtype in struct_types:  
            return struct_types[dtype]['size']
        return None

if __name__ == '__main__':
    #
    #  Give port number or other through command line arguments
    #         
    hdw_server = HardwareServer()
    hdw_server.run()



#!/usr/bin/env python3 
# -*- coding: utf-8 -*-
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%

"""
A minimal front end to the Docutils Publisher, producing HTML.
"""

import sys, re

DEBUG = 0

try:
    from docutils.core      import publish_cmdline, publish_file, Publisher
    from docutils.writers   import manpage
    from docutils           import nodes
    from docutils.nodes     import reprunicode
    from docutils.writers.manpage import Writer, Translator, Table
    from docutils.core import publish_string
except ImportError:
    print("Cannot import docutils")
    sys.exit(1)

if "-check" in sys.argv:
    sys.exit(0)

DEFINITION_LIST_INDENT = 5
FIELD_LIST_INDENT = 7
OPTION_LIST_INDENT = 7
BLOCKQUOTE_INDENT = 4
LITERALBLOCK_INDENT = 3


def debug(fn):
    if DEBUG:
       def wrapped(self, node):
          self.body.append('.\" <ENTER : %s >\n' % fn.__name__)
          return fn(self, node)
    else:
       def wrapped(self, node):
          return fn(self, node)
    return wrapped

class SpecMANTranslator(Translator):

   document_start = "spec reStructuredText documentation"

   def __init__(self,document):

       Translator.__init__(self,document)

       self.restro = re.compile(r"\*\*(?!\ )(?P<emph>[\\@=\-()\:#+/a-zA-Z0-9_\ ]+)(?<!\ )\*\*")
       self.remph  = re.compile(r"\*(?!\ )(?P<emph>[\\@=\-()\:#+/a-zA-Z0-9_\ ]+)(?<!\ )\*")
       self.keyw   = re.compile(r"(?P<keyw>[a-zA-Z0-9_]+)\s?\:\s?(?P<value>.*?)$")
 
       self.deflist_level = 0

       self.hanging    = 0
       self.inlistitem = 0

       self.intitle    = 0
       self.insection  = 0
       self.insubsect  = 0
       self.section_visited = 0

       self.prevent_nestfont = 0
       self.literalblock = 0

       self.doctitle = ""
       self.docsubtitle = ""
 
       self.defs = {
                'indent' : ('.RS %dm\n', '.RE\n'),

                'definition_list' : ('', ''),
                'definition_list_item' : ('.HP %dm\n', ''),
                'field_name' : ('.HP 5m\n', '\n'),
                'literal' : ('\\fB', '\\fP'),
                'literal_block' : ('.sp\n.nf\n.ft CB\n', '\n.ft P\n.fi\n'),

                'option_list_item' : ('.HP 5m\n', ''),

                'reference' : (r'\fI\%', r'\fP'),
                'emphasis': ('\\fB', '\\fP'),
                'strong' : ('\\fB', '\\fP'),
                'term' : ('\n', '\n'),
                'title_reference' : ('\\fI', '\\fP'),

                'topic-title' : ('.SH ',),
                'sidebar-title' : ('.SH ',),

                'problematic' : ('\n.nf\n', '\n.fi\n'),
                    }

       self._docinfo = {
                "title" : "", "title_upper": "",
                "subtitle" : "",
                "manual_section" : "", "manual_group" : "",
                "author" : [],
                "date" : "",
                "copyright" : "",
                "version" : "",
                "versioninfo" : "",
            }


   @debug
   def visit_versioninfo(self,node):
       self._docinfo["versioninfo"] =  " ".join( node.content )

   @debug
   def depart_versioninfo(self,node):
       pass

   @debug
   def visit_document(self,node):
       Translator.visit_document(self,node)
       self.doctitle = node.get('title')

   @debug
   def depart_document(self,node):
       pass

   @debug
   def visit_title(self, node):

       self.deflist_level = 0
       self.hanging       = 0

       if isinstance(node.parent, nodes.topic):
           self.body.append(self.defs['topic-title'][0])
       elif isinstance(node.parent, nodes.sidebar):
           self.body.append(self.defs['sidebar-title'][0])
       elif isinstance(node.parent, nodes.admonition):
           self.body.append('.IP "')

       elif self.section_level == 0:
           self._docinfo['title_upper'] = node.astext().upper()
           self.intitle = 1
       elif self.section_level == 1:
           self.insection = 1
       else:
           self.insubsect = 1

       self._keepbody = self.body
       self.body = []


   def depart_title(self, node):

        title = ''.join(self.body)
        self.body = self._keepbody

        if isinstance(node.parent, nodes.admonition):
            self.body.append('"')

        if self.insection:
            self.insection = 0
            title = re.sub(r'"',r'\\(dq',title)
            self.body.append('.SH \"%s\"\n.rs\n' % title )
           
        if self.insubsect:
            self.insubsect = 0
            title = re.sub(r'"',r'\\(dq',title)
            self.body.append('.SH \"%s\"\n' % title )

        if self.intitle:
            self.intitle = 0
            self._docinfo["title"] = title

        # if we miss level 0. there is a problem
        if not self.doctitle and self.section_level == 1 :
           self.doctitle = "NO TITLE"
           self._docinfo["title"] = title

        if not self.docsubtitle and self.section_level == 2 :
           self.docsubtitle = "Document with no title. Please provide one"
           self._docinfo["subtitle"] = title


   @debug
   def visit_subtitle(self, node):
        if isinstance(node.parent, nodes.sidebar):
            self.body.append(self.defs['strong'][0])
        elif isinstance(node.parent, nodes.document):
            # self.visit_docinfo_item(node, 'subtitle')
            # to catch and process subtitle we need to skip docinfo call
            self.subtitle_raw = node.astext()
            self.insubtitle = 1
            self._keepbody = self.body
            self.body = []

            self._docinfo_keys.append('subtitle')
        elif isinstance(node.parent, nodes.section):
            self.body.append(self.defs['strong'][0])

   def depart_subtitle(self, node):
        # document subtitle calls SkipNode
        self.subtitle = ''.join( self.body )
        self._docinfo['subtitle'] = self.subtitle
        self.docsubtitle = self.subtitle
        self.body = self._keepbody
        self.body.append(self.defs['strong'][1]+'\n.PP\n')

   def visit_comment(self, node, sub=re.compile('-(?=-)').sub):
       ## search for keywords in comments only if we have not visited
       # a section yet
       if self.section_visited == 0:
            mat = self.keyw.search( node.astext() )
            if mat:
                 keyw  = mat.group("keyw")
                 value = mat.group("value")
                 self._docinfo[keyw] = value

       self.body.append(self.comment(node.astext()))
       raise nodes.SkipNode

   def header(self):
       if self._docinfo['title']:
           self._docinfo['title_upper'] = self._docinfo['title']

       # Allow for basic formatting in versioninfo

       self.prevent_nestfont = 1
       self._docinfo['versioninfo'] = self.process_Text( self._docinfo['versioninfo'] )
       self.prevent_nestfont = 0

       tmpl  = (".TH %(versioninfo)s\n"
                ".ds HF R\n"
                ".na\n"      #  no justification
                ".hy 0\n"    #  do not hyphenate words
                ".SH NAME\n"
               "%(title)s \- %(subtitle)s\n")

       return tmpl % self._docinfo

   def visit_paragraph(self, node):
       self.ensure_eol()
       if not self.first_child(node):
           self.body.append('.sp\n')

   def first_child(self, node):
       first = isinstance(node.parent[0], nodes.label) # skip label
       for child in node.parent.children[first:]:
           if isinstance(child, nodes.Invisible):
               continue
           if child is node:
               return 1
           break
       return 0

   @debug
   def visit_section(self,node):
       self.section_visited += 1
       Translator.visit_section(self,node)

   def myindent(self, by=5):
        # if we are in a section ".SH" there already is a .RS
        self.body.append(self.defs['indent'][0] % by)

   def indent(self,by=0):
       pass

   def mydedent(self, by=5):
        self.body.append(self.defs['indent'][1])
   def dedent(self,by=0):
       pass


   """
    Change from original docutils manpage Writer 
    is that here we use our own Table class
   """
   def visit_table(self, node):
      self._active_table = SpecTable()

   def depart_table(self, node):
      self.ensure_eol()
      self.body.extend(self._active_table.as_list())
      self._active_table = None

   """
    This solves the issues:
    - allow for emphasis inside literal blocks (necessary for spec and c-plot
      function parameters.  This functionality should actually be part of the parser
      but it would affect too many classes and compromise later compatibility
   """ 
   @debug
   def visit_literal(self, node):
        """
        Inside literal will replace a string of letter, numbers and hyphens surrounded by * to be emphasized
        """ 

        """Process text to get away extra formatting in literals (extra formatting is used in HTML, but not supported here)"""

        self.prevent_nestfont = 1
        text = self.process_Text( node.astext() )

        self.body.append(self.defs['literal'][0])
        self.body.append(text)
        self.body.append(self.defs['literal'][1])

        # Content already processed:
        raise nodes.SkipNode

   @debug
   def depart_literal(self, node):
        self.prevent_nestfont = 0

   def visit_Text(self, node):
        text = self.process_Text( node.astext() )
        self.body.append(text)

   def process_Text(self, text):
        # reserve escaped backlashes (in text as \\)
        text = re.sub("\\\\\\\\","&7&",text)

        # reserve escaped asterisks (in text as \*)
        text = re.sub("\\\\\*", "@=@",text)

        text = text.replace('\\','\\e')

        replace_pairs = [
	    (u'-', r'\-'),
	    (u'\'', r'\(aq'),
	    (u'`', r'\(ga'),
            ]
        for (in_char, out_markup) in replace_pairs:
            text = text.replace(in_char, out_markup)

        # unicode
        text = self.deunicode(text)
        if self._in_literal:
            # prevent interpretation of "." at line start
            if text[0] == '.':
                text = '\\&' + text
            text = text.replace('\n.', '\n\\&.')


        if self.prevent_nestfont:
           text = self.restro.sub(r"\g<emph>",text )
           text = self.remph.sub(r"\g<emph>",text )
        else:
           text = self.restro.sub(r"\\fB\g<emph>\\fP",text )
           text = self.remph.sub(r"\\fI\g<emph>\\fP",text )

        # restore escaped asterisks as normal asterisks
        text = re.sub("@=@", "*", text)
        # restore escaped backlashes as normal backslashes
        text = re.sub("&7&", "\\\\e",  text)


        return text
        
   def depart_Text(self, node):
       pass

   @debug
   def visit_bullet_list(self, node):
        self.list_start(node)

   @debug
   def depart_bullet_list(self, node):
        self.list_end()

   @debug
   def visit_literal_block0(self, node):

        """Process text to get away extra formatting in literals (extra formatting is used in HTML, but not supported here)"""
        """
          Indent the whole literal block by two spaces to avoid characters in first column 
          to be interpreted by nroff
        """
        #text = re.sub("\n","\n  ",text )
        #text = "  "+text

        self.body.append(self.defs['literal_block'][0])
        self.body.append( text )
        self.body.append(self.defs['literal_block'][1])

        # Content already processed:
        raise nodes.SkipNode

   @debug
   def depart_literal_block0(self, node):
        pass

   def visit_literal_block(self, node):
        if self.inlistitem and self.hanging == self.deflist_level: 
           self.hanging += 1
           self.myindent(DEFINITION_LIST_INDENT)
        self.myindent(LITERALBLOCK_INDENT)
        self.body.append(self.defs['literal_block'][0])
        self._in_literal = True
        self.prevent_nestfont = 1
        self.literalblock = 1

   def depart_literal_block(self, node):
        self._in_literal = False
        self.prevent_nestfont = 0
        self.literalblock = 0
        self.body.append(self.defs['literal_block'][1])
        self.mydedent()

   @debug
   def visit_block_quote(self, node):
        self.body.append(".sp\n")
        self.prevent_nestfont = 1
        if self.inlistitem and self.hanging == self.deflist_level: 
           self.hanging += 1
           self.myindent(DEFINITION_LIST_INDENT)
        self.myindent(BLOCKQUOTE_INDENT)

   @debug
   def depart_block_quote(self, node):
        self.prevent_nestfont = 0
        self.mydedent()

   @debug
   def visit_definition_list_item(self, node):
        self.inlistitem = 1
        self.body.append(self.defs['definition_list_item'][0] % DEFINITION_LIST_INDENT)
        self.body.append('."')

   @debug
   def depart_definition_list_item(self, node):
        if self.hanging > self.deflist_level: 
            self.mydedent()
            self.hanging -= 1
        self.inlistitem = 0
        self.body.append(self.defs['definition_list_item'][1])

   @debug
   def visit_definition_list(self, node):
       self.deflist_level += 1
       self.hanging       += 1
       if self.deflist_level > 1:
          self.myindent(DEFINITION_LIST_INDENT)
          self.body.append('.sp\n.PD 0.2v\n')
       # set a reference with no indent to return to
       self.myindent(0)

   @debug
   def depart_definition_list(self, node):

       if self.deflist_level == 2:
          self.body.append('.PD\n')

       self.mydedent()
       self.deflist_level -=1

   @debug
   def visit_definition(self,node):
       self.body.append('\- \n')

   @debug
   def visit_enumerated_list(self, node):
        self.list_start(node)

   @debug
   def depart_enumerated_list(self, node):
        self.list_end()

   @debug
   def visit_field(self, node):
        pass

   @debug
   def depart_field(self, node):
        pass

   @debug
   def visit_field_body(self, node):
        if self._in_docinfo:
            name_normalized = self._field_name.lower().replace(" ","_")
            self._docinfo_names[name_normalized] = self._field_name
            self.visit_docinfo_item(node, name_normalized)
            raise nodes.SkipNode

   @debug
   def depart_field_body(self, node):
        pass

   @debug
   def visit_field_list(self, node):
        self.myindent(FIELD_LIST_INDENT)

   @debug
   def depart_field_list(self, node):
        self.mydedent()

   @debug
   def visit_field_name(self, node):
        if self._in_docinfo:
            self._field_name = node.astext()
            raise nodes.SkipNode
        else:
            self.body.append(self.defs['field_name'][0])

   @debug
   def depart_field_name(self, node):
        self.body.append(self.defs['field_name'][1])

   @debug
   def visit_figure(self, node):
        self.myindent(2.5)
        self.myindent(0)

   @debug
   def depart_figure(self, node):
        self.mydedent()
        self.mydedent()


    # WHAT should we use .INDENT, .UNINDENT ?
   @debug
   def visit_line_block(self, node):
        self._line_block += 1
        if self._line_block == 1:
            # TODO: separate inline blocks from previous paragraphs
            # see http://hg.intevation.org/mercurial/crew/rev/9c142ed9c405
            # self.body.append('.sp\n')
            # but it does not work for me.
            self.body.append('.nf\n')
        else:
            self.body.append('.in +2\n')

   @debug
   def depart_line_block(self, node):
        self._line_block -= 1
        if self._line_block == 0:
            self.body.append('.fi\n')
            self.body.append('.sp\n')
        else:
            self.body.append('.in -2\n')

   @debug
   def visit_option_list(self, node):
        self.myindent(OPTION_LIST_INDENT)

   @debug
   def depart_option_list(self, node):
        self.mydedent()

   def depart_paragraph(self, node):
        self.body.append('\n')

   @debug
   def visit_system_message(self,node,name=None):
       raise nodes.SkipNode

class SpecTable(Table):
     """
     Redefines the way a table is drawn from the original 
     doctutils manpage writer.

     Changes:  set no border
     """

     def __init__(self):
        self._rows = []
        self._options = ['']
        self._tab_char = '\t'
        self._coldefs = []

     def as_list(self):
        text = ['.TS\n']
        text.append(' '.join(self._options) + ';\n')
        text.append('%s .\n' % (' '.join(self._coldefs)))
        for row in self._rows:
            # row = array of cells. cell = array of lines.
            #text.append('_\n')       # line above
            #text.append('T{\n')
            for i in range(len(row)):
                cell = row[i]
                self._minimize_cell(cell)
                text.extend(cell)
                #if not text[-1].endswith('\n'):
                #    text[-1] += '\n'
                if i < len(row)-1:
                    text.append(self._tab_char)
                else:
                    text.append('\n')
                #if i < len(row)-1:
                #    text.append('T}'+self._tab_char+'T{\n')
                #else:
                #    text.append('T}\n')
        #text.append('_\n')
        text.append('.TE\n')
        return text


class SpecMANWriter(Writer):


   def __init__(self):
      Writer.__init__(self)
      self.translator_class = SpecMANTranslator
      self.template_str = ""
      self.extraclass   = None

   def setTemplate(self, templ_file):
      self.template_file = templ_file
      self.template_str = open(self.template_file).read()

   def apply_template(self):

      if not self.template_str:
         print("NO TEMPLATE")
         raise "NoTemplate"

      subs = self.interpolation_dict()

      subs['description'] = self.visitor.doctitle

      return self.template_str % subs

   def translate(self):
      self.visitor = self.translator_class(self.document)
      self.document.walkabout(self.visitor)

      self.doctitle    = self.visitor.doctitle

      try:
         self.docsubtitle = self.visitor.docsubtitle
      except:
         self.docsubtitle = "NO SUBTITLE"

      self.output = self.visitor.astext()


def process(filename, outfile=None):
    buf = open(filename).read()
    pub_str = publish_string(buf, writer=SpecMANWriter(), settings_overrides={'output_encoding': 'unicode'})

    if not outfile:
       fd = sys.stdout
    else:
       fd = open(outfile,"w")

    fd.write(pub_str)

if __name__ == '__main__':

    import sys


    if "-d" in sys.argv:
        idx = sys.argv.index("-d")
        sys.argv.pop(idx)
        DEBUG = 1

    if len(sys.argv) < 2:
        print("Usage: %s filename [outfile]", sys.argv[0])
        sys.exit(0)

    filename = sys.argv[1]
    if len(sys.argv) >= 3:
        outfile = sys.argv[2]
    else:
        outfile = None

    process(filename, outfile)



#   %W%  %G% CSS
#
#   "spec" Release %R%
#
"""
class DEigerClient provides an interface to the EIGER API

Author: Volker Pilipp
Contact: support@dectris.com
Version: 1.0
Date: 13/11/2014
Copyright See General Terms and Conditions (GTC) on http://www.dectris.com

"""
import threading

import sys
import time
import base64
import os.path

if sys.version_info[0] >= 3:
    import http.client as http
    from urllib.request import urlopen
    from io import IOBase
else:
    import httplib as http
    from urllib2 import urlopen

import json
import re
import fnmatch
import shutil

from pyspec.hardware.server import log

Version = '1.5.0'

TRIGGER_NO_MAX = 1000000
#MAX_TIMEOUT = 60 # sixty seconds
MAX_TIMEOUT = 10*60 # ten minutes
# MAX_TIMEOUT = 24*3600 # one full day

class CommandHandlerThread(threading.Thread):

    commandTimeout = MAX_TIMEOUT

    def __init__(self, host, port, url, command, data, mimeType):
        log.log(2,"creating command thread for command %s" % command)
        self._connection = http.HTTPConnection(host,port,timeout= self.commandTimeout)
        self.url = url
        self.cmd_name = command
        self.data = data
        self.mimeType = mimeType 
        threading.Thread.__init__(self)
        self.answer = None

    def getCommandName(self):
        return self.cmd_name

    def getCommandAnswer(self):
        return self.answer

    def run(self):
        self.answer = self.run_command(self.cmd_name, self.url)

        if self.answer is not None:
            return self.answer
        else:
            return self.error()

    def run_command(self, cmdname, url):
        log.log(2,"command %s - (%s) is running" % (self.cmd_name, cmdname))

        headers = {}
        headers['Content-type'] = self.mimeType
        
        self._connection.request('PUT', url, body = self.data, headers = headers)
        response = self._connection.getresponse()
        status = response.status
        data = response.read()

        if status == 200:
            data_l = json.loads(data)
            if isinstance(data_l,dict) and cmdname in ['arm', 'trigger']:
                log.log(2, "answer to %s is : %s" % (cmdname, str(data_l)))
                if 'sequence_id' in data_l.keys():
                    answer = data_l['sequence id']
                else:
                    log.log(3, "answer to %s contains keys: %s" % (cmdname, data_l.keys()))
                    answer = data_l
            else: 
                answer = data_l
            log.log(2, "response from %s command :  %s" % (cmdname, str(answer)))
            return answer
        else:
            log.log(1, "cannot get response from %s command - status: %s" % (cmdname,status))
            return None

    def error(self):
        self.answer = None
        return '.error.'

class StartAcquisitionThread(CommandHandlerThread):

    cmd_name = 'startacq'

    def __init__(self, host, port, urls):

        CommandHandlerThread.__init__(self, host, port, None, 'startacq', data='', mimeType='text/html')

        self.urls = urls

        self.trigger_mode = None
        self.ntrigger = None
        self.count_time = None

    def set_arm_needed(self, needed):
        self.arm_needed = needed

    def set_trigger_mode(self, trigger_mode):
        self.trigger_mode = trigger_mode

    def set_ntrigger(self, value):
        self.ntrigger = value

    def set_count_time(self, value):
        self.count_time = value

    def set_par(self, name, value):
        headers = {}
        headers['Content-type'] = 'application/json; charset=utf-8'
        data = json.dumps({'value':value})
        log.log(2,"start thread - setting %s to %s" % (name,value))
        url = self.urls['config'] + '{0}'.format(name)

        numberOfTries = 0
        response = None

        try:
            while response is None:
                try:
                    self._connection.request('PUT',url, body = data, headers = headers)
                    response = self._connection.getresponse()
                except Exception as e:
                    numberOfTries += 1
                    if numberOfTries == 50:
                        log.log(2,"Terminate after {0} tries\n".format(numberOfTries))
                        raise e
                    log.log(2,"Failed to connect to host. Retrying\n")
                    self._connection = http.HTTPConnection(self._host,self._port, timeout = self._connectionTimeout)
                    continue

            status = response.status
            reason = response.reason
            log.log(2,"response arrived status:%s / reason. %s" % (status,reason))
            data = response.read()
        except:
            log.log(2,'setpar error')
        finally:
            pass

        log.log(2,"setting %s returns %s " % (name,str(data)))

    def send_cmd(self, cmd):
        log.log(2, "start thread - starting command %s" % cmd)
        url = self.urls['command'] + '{0}'.format(cmd)
        aret = self.run_command(cmd, url)
        return aret

    def run(self):
        try:
            if self.trigger_mode in ['ints', 'inte']:
                if self.arm_needed: 
                    if self.ntrigger == 1:
                        log.log(2,"programming big trigger no")
                        self.set_par("ntrigger",TRIGGER_NO_MAX)
                    elif self.ntrigger == TRIGGER_NO_MAX:
                        log.log(3,"ntrigger is big")
                    else:
                        log.log(2,"programming ntrigger to %s" % self.ntrigger)
                        self.set_par(self.ntrigger)

                    if self.count_time is not None:
                        self.set_par("count_time", self.count_time)
                        self.set_par("frame_time", self.count_time+0.00002)

                    aret = self.send_cmd("arm")
                    log.log(2, "arm command finished. returned: %s" % aret)
    
                    if aret is None:
                        return self.error()

                else:
                    aret = None

                tret = self.send_cmd("trigger")
                log.log(2, "trigger command finished. returned: %s" % aret)

                if tret is not None:
                    if aret is None:
                        self.answer = tret
                        log.log(2, "startacq finished. returning t -  %s " % tret)
                        return
                    else:
                        log.log(2, "startacq finished. returning a -  %s " % aret)
                        self.answer = aret
                        return

            else:  # external trigger
                aret = self.send_cmd("arm")
                log.log(2, "arm command finished. returned: %s" % aret)
                self.set_par("ntrigger", self.ntrigger)
    
                if aret is None:
                    return self.error()

        except BaseException as exc:
            import traceback
            log.log(1,"error starting acquisition %s" % traceback.format_exc())
            
        return self.error()

class FileWriterThread(threading.Thread):
   
    commandTimeout = MAX_TIMEOUT

    def __init__(self, host, port, urlprefix, filename, targetDir):
        self.target_path = os.path.join(targetDir,filename)
        self.url = 'http://{0}:{1}/{2}data/{3}'.format(host, port, urlprefix, filename)
        threading.Thread.__init__(self)
    
    def run(self):
        req = urlopen(self.url, timeout = self.commandTimeout)

        try:
            with open(self.target_path, 'wb') as fp:
                log.log(2,'writing ' + self.target_path)
                shutil.copyfileobj(req, fp, 512*1024)
                assert os.access(self.target_path,os.R_OK)
            return "ok"
        except:
            import traceback
            log.log(1, 'cannot write ' + self.target_path)
            log.log(1,  traceback.format_exc() )
            return ".error."

# noinspection PyInterpreter
class DEigerClient(object):
    """
    class DEigerClient provides a low level interface to the EIGER API
    """

    def __init__(self, host = '127.0.0.1', port = 80, verbose = False, urlPrefix = None, user = None):
        """
        Create a client object to talk to the EIGER API.
        Args:
            host: hostname of the detector computer
            port: port usually 80 (http)
            verbose: bool value
            urlPrefix: String prepended to the urls. Should be None. Added for future convenience.
            user: "username:password". Should be None. Added for future convenience.
        """
        super(DEigerClient,self).__init__()
        self._host = host
        self._port = port
        self._version = Version
        self._verbose = verbose
        self._urlPrefix = ""
        self._user = None
        self._connectionTimeout = MAX_TIMEOUT

        log.log(2, "DEigerClient created - host: %s / port: %s (verbose=%s)" % (self._host, self._port, self._verbose))

        #pingret = os.system("ping -c 1 -w 1 -W 10 %s > /dev/null 2>&1" % self._host)
        #if pingret != 0:
            #raise BaseException("Cannot access DCU")

        try:
            self._connection = http.HTTPConnection(self._host,self._port, timeout = self._connectionTimeout)
        except BaseException as exc:
            log.log(1,"Exception while connecting to Eiger DCU %s" % str(exc))
            raise(e)

        self._serializer = None
        
        self.detector_command = None
        self.cmd_answer = {}

        self.read_thread = None
        self.read_value = {}
        self.setUrlPrefix(urlPrefix)
        self.setUser(user)
        
    def serializer(self):
        """
        The serializer object shall have the methods loads(string) and dumps(obj), which load
        the string from json into a python object or store a python object into a json string
        """
        return self._serializer
        
    def setSerializer(self,serializer):
        """
        Set an explicit serializer object that converts native python objects to json string and vice versa.
        The serializer object shall have the methods loads(string) and dumps(obj), which load
        the string from json into a python object or store a python object into a json string
        """
        self._serializer = serializer

    def setVerbose(self,verbose):
        """ Switch verbose mode on and off.
        Args:
            verbose: bool value
        """
        self._verbose = bool(verbose)
        log.log(2, "DEigerClient verbose level set to %s" % (verbose))
        
    def setConnectionTimeout(self, timeout):
        """
        If DEigerClient has not received an reply from EIGER after 
        timeout seconds, the request is aborted. timeout should be at 
        least as long as the triggering command takes.
        Args:
            timeout timeout in seconds
        """
        self._connectionTimeout = timeout
        self._connection = http.HTTPConnection(self._host,self._port, timeout = self._connectionTimeout)
        
    def setUrlPrefix(self, urlPrefix):
        """Set url prefix, which is the string that is prepended to the 
        urls. There is usually no need to call the command explicitly.
        Args:
           urlPrefix: String
        """
        if urlPrefix is None:
            self._urlPrefix = ""
        else:
            self._urlPrefix = str(urlPrefix)
            if len(self._urlPrefix) > 0 and self._urlPrefix[-1] != "/":
                self._urlPrefix += "/"
                
    def setUser(self, user):
        """
        Set username and password for basic authentication.
        There is usually no need to call the command explicitly.
        Args:
           user: String of the form username:password 
        """
        if user is None:
            self._user = None
        else:
            self._user = base64.encodestring(user).replace('\n', '')
       
    def set_verbose(self, verbose):
        self._verbose = verbose

    def set_version(self, version):
        self._log("eigerclient version set to %s" % version)
        self._version = version

    def version(self,module = 'detector'):
        """
        Get version of a api module (i.e. 'detector', 'filewriter')
        Args:
            module: 'detector' or 'filewriter'
        """
        return self._getRequest(url = '/{0}{1}/api/version/'.format(self._urlPrefix,module))['value']

    def sendSystemCommand(self, command):
        """
        Sending command "restart" restarts the SIMPLON API on the EIGER control unit
        """
        return self._putRequest(self._url('system','command',command), dataType = 'native', data = None)

    def listDetectorConfigParams(self):
        """Get list of all detector configuration parameters (param arg of configuration() and setConfiguration()).
        Convenience function, that does detectorConfig(param = 'keys')
        Returns:
            List of parameters.
        """
        return self.detectorConfig('keys')

    def detectorConfig(self,param = None, dataType = None):
        """Get detector configuration parameter
        Args:
            param: query the configuration parameter param, if None get full configuration, if 'keys' get all configuration parameters.
            dataType: None (= 'native'), 'native' ( return native python object) or 'tif' (return tif data).
        Returns:
            If param is None get configuration, if param is 'keys' return list of all parameters, else return the value of
            the parameter. If dataType is 'native' a dictionary is returned that may contain the keys: value, min, max,
            allowed_values, unit, value_type and access_mode. If dataType is 'tif', tiff formated data is returned as a python
            string.
        """
        return self._getRequest(self._url('detector','config',param),dataType)

    def setDetectorConfig(self, param, value, dataType = None):
        """
        Set detector configuration parameter param.
        Args:
            param: Parameter
            value: Value to set. If dataType is 'tif' value may be a string containing the tiff data or
                   a file object pointing to a tiff file.
            dataType: None, 'native' or 'tif'. If None, the data type is auto determined. If 'native' value
                      may be a native python object (e.g. int, float, str), if 'tif' value shell contain a
                      tif file (python string or file object to tif file).
        Returns:
            List of changed parameters.
        """
        return self._putRequest(self._url('detector','config',param), dataType, value)

    def setDetectorConfigMultiple(self,*params):
        """
        Convenience function that calls setDetectorConfig(param,value,dataType = None) for
        every pair param, value in *params.
        Args:
            *params: List of successive params of the form param0, value0, param1, value1, ...
                     The parameters are set in the same order they appear in *params.
        Returns:
            List of changed parameters.
        """
        changeList = []
        p = None
        for x in params:
            if p is None:
                p = x
            else:
                data = x
                changeList += self.setDetectorConfig(param = p, value = data, dataType = None)
                p = None
        return list(set(changeList))

    def listDetectorCommands(self):
        """
        Get list of all commands that may be sent to EIGER via sendDetectorCommand().
        Returns:
            List of commands
        """
        return self._getRequest(self._url('detector','command','keys'))

    def sendDetectorCommand(self,  command, parameter = None):
        """
        Send command to EIGER. The list of all available commands is obtained via listCommands().
        Args:
            command: Detector command 
            parameter: Call command with parameter. If command = "trigger" a float parameter may be passed
        Returns:
            The commands 'arm' and 'trigger' return a dictionary containing 'sequence id'.
        """
        if command not in ['cancel', 'abort']:
            self.wait_command_end()

        url = self._url('detector','command',command)
        data, mimeType = self._prepareData(parameter,'native')
        self.detector_command = CommandHandlerThread(self._host, self._port, url, command, data, mimeType)
        self.startDetectorCommand()
        return 

    def startAcquisition(self, need_arm = True, trigger_mode='ints', ntrigger=1, count_time=None):
        if self.isExecutingCommand():
            log.log(1,"cannot start acquisition while a command is running")
            return False

        urls = {
                'config': self._url('detector','config'),
                'command': self._url('detector','command'),
               }

        self.detector_command = StartAcquisitionThread(
                                    self._host, self._port, urls)

        self.detector_command.set_arm_needed(need_arm)
        self.detector_command.set_trigger_mode(trigger_mode)
        self.detector_command.set_ntrigger(ntrigger)
        self.detector_command.set_count_time(count_time)

        self.startDetectorCommand()

        return True

    def startDetectorCommand(self):
        cmd_name = self.detector_command.getCommandName()
        self.cmd_answer[cmd_name] = None
        self.detector_command.start()

    def wait_command_end(self):
        while self.isExecutingCommand():
            self._log('waiting for running command (%s) to finish ', self.detector_command.getCommandName())
            time.sleep(0.1)

    def updateCommandStatus(self):
        if self.detector_command is None:
            return 
        
        cmd_alive = self.detector_command.is_alive()

        if cmd_alive is False:
            try:
                cmd_name = self.detector_command.getCommandName()
                ret_value = self.detector_command.getCommandAnswer()
                if cmd_name == 'check_connections':
                    try:
                        ans = ret_value['value']    
                        ret_value = ""
                        for conn in ans: 
                            ret_value += '{name}={up},'.format(**conn)
                    except:
                        ret_value = "cannot parse result"
                else:
                    ret_value = str(ret_value)
                self.cmd_answer[cmd_name] = ret_value
                log.log(4,"command %s finished execution. answer is: %s" % (cmd_name, ret_value))
            except BaseException as e:
                log.log(1,"error update command: %s" % str(e))

            # self.detector_command = None


    def isExecutingCommand(self):
        self.updateCommandStatus()
        if self.detector_command is not None:
            return self.detector_command.is_alive()
        else:
            return False

    def getSequenceID(self):
        return self.cmd_answer.get('startacq', None)

    def getCommandName(self):
        if self.detector_command is None:
            return "none"
        
        return self.detector_command.getCommandName()

    def getAnswer(self, command=None):
        if self.detector_command is None:
            return "none"
        
        latest_command = self.getCommandName()

        if command is None or command == latest_command: 
            if self.detector_command.is_alive():
                return "busy"

        if command is None:
            command = latest_command
         
        return self.cmd_answer.get(command,"none")

    def detectorStatus(self, param = 'keys'):
        """Get detector status information
        Args:
            param: query the status parameter param, if 'keys' get all status parameters.
        Returns:
            If param is None get configuration, if param is 'keys' return list of all parameters, else return dictionary
            that may contain the keys: value, value_type, unit, time, state, critical_limits, critical_values
        """
        if param == 'state' and self.isExecutingCommand():
            return {'value':'running'}

        return self._getRequest(self._url('detector','status',parameter = param))


    def fileWriterConfig(self,param = 'keys'):
        """Get filewriter configuration parameter
        Args:
            param: query the configuration parameter param, if 'keys' get all configuration parameters.
        Returns:
            If param is None get configuration, if param is 'keys' return list of all parameters, else return dictionary
            that may contain the keys: value, min, max, allowed_values, unit, value_type and access_mode
        """
        return self._getRequest(self._url('filewriter','config',parameter = param))

    def setFileWriterConfig(self,param,value):
        """
        Set file writer configuration parameter param.
        Args:
            param: parameter
            value: value to set
        Returns:
            List of changed parameters.
        """
        return self._putRequest(self._url('filewriter','config',parameter = param), dataType = 'native', data = value)

    def sendFileWriterCommand(self, command):
        """
        Send filewriter command to EIGER.
        Args:
            command: Command to send (up to now only "clear")
        Returns:
            Empty string
        """
        return self._putRequest(self._url("filewriter","command",parameter = command), dataType = "native")


    def fileWriterStatus(self,param = 'keys'):
        """Get filewriter status information
        Args:
            param: query the status parameter param, if 'keys' get all status parameters.
        Returns:
            If param is None get configuration, if param is 'keys' return list of all parameters, else return dictionary
            that may contain the keys: value, value_type, unit, time, state, critical_limits, critical_values
        """
        return self._getRequest(self._url('filewriter','status',parameter = param))

    def fileWriterFiles(self, filename = None, method = 'GET'):
        """
        Obtain file from detector.
        Args:
             filename: Name of file on the detector side. If None return list of available files
             method: 'GET' (get the content of the file) or 'DELETE' (delete file from server)
        Returns:
            List of available files if 'filename' is None,
            else if method is 'GET' the content of the file.
        """
        if method == 'GET':
            if filename is None:
                return self._getRequest(self._url('filewriter','files'))

            return self._getRequest(url = '/{0}data/{1}'.format(self._urlPrefix, filename), dataType = 'hdf5')
        elif method == 'DELETE':
            return self._delRequest(url = '/{0}data/{1}'.format(self._urlPrefix,filename))
        else:
            raise RuntimeError('Unknown method {0}'.format(method))

    def fileWriterSave(self,filename,targetDir,regex = False):
        """
        Saves filename in targetDir. If regex is True, filename is considered to be a regular expression.
        Save all files that match filename
        Args:
            filename: Name of source file, may contain the wildcards '*' and '?' or regular expressions
            targetDir: Directory, where to store the files
        """
        if regex:
            pattern = re.compile(filename)
            for f in self.fileWriterFiles():
                if pattern.match(f):
                    self.fileWriterSave(f,targetDir)
        elif any([ c in filename for c in ['*','?','[',']'] ] ):
            for f in self.fileWriterFiles():
                if fnmatch.fnmatch(f,filename):
                    self.fileWriterSave(f,targetDir)
        else:
            worker = FileWriterThread(self._host, self._port, self._urlPrefix, filename, targetDir)
            worker.start()
        return

    def monitorConfig(self,param = 'keys'):
        """Get monitor configuration parameter
        Args:
            param: query the configuration parameter param, if 'keys' get all configuration parameters.
        Returns:
            If param is 'keys' return list of all parameters, else return dictionary
            that may contain the keys: value, min, max, allowed_values, unit, value_type and access_mode
        """
        return self._getRequest(self._url('monitor','config',parameter = param))

    def setMonitorConfig(self,param,value):
        """
        Set monitor configuration parameter param.
        Args:
            param: parameter
            value: value to set
        Returns:
            List of changed parameters.
        """
        return self._putRequest(self._url('monitor','config',parameter = param), dataType = 'native', data = value)

    def monitorImages(self, param = None):
        """
        Obtain file from detector.
        Args:
             param: Either None (return list of available frames) or "monitor" (return latest frame),
                    "next"  (next image from buffer) or tuple(sequence id, image id) (return specific image)
        Returns:
            List of available frames (param = None) or tiff content of image file (param = "next", "monitor", (seqId,imgId))
        """
        if param is None:
            return self._getRequest(self._url('monitor','images',parameter = None) )
        elif param == "next":
            return self._getRequest(self._url('monitor',"images", parameter = "next"), dataType = "tif")
        elif param == "monitor":
            return self._getRequest(self._url('monitor','images',parameter = "monitor"), dataType = "tif")
        else:
            try:
                if len(param) == 2:
                    seqId = int(param[0])
                    imgId = int(param[1])
                    return self._getRequest(self._url('monitor',"images", parameter = "{0}/{1}".format(seqId,imgId) ), dataType = 'tif')
                elif len(param) == 3:
                    seqId = int(param[0])
                    imgId = int(param[1])
                    threshId = int(param[2])
                    return self._getRequest(self._url('monitor',"images", parameter = "{0}/{1}/{2}".format(seqId,imgId,threshId) ), dataType = 'tif')
            except (TypeError, ValueError):
                pass
        raise RuntimeError('Invalid parameter {0}'.format(param))

    def monitorSave(self, param, path):
        """
        Save frame to path as tiff file.
        Args:
            param: Either None (return list of available frames) or "monitor" (return latest frame),
                   "next"  (next image from buffer) or tuple(sequence id, image id) (return specific image)        
        Returns:
            None
        """
        data = None
        if param in ["next","monitor"]:
            data = self.monitorImages(param)
        else :
            try:
                int(param[0])
                int(param[1])
                data = self.monitorImages(param)
            except (TypeError, ValueError):
                pass
        if data is None:
            raise RuntimeError('Invalid parameter {0}'.format(param))
        else:
            with open(path,'wb') as f:
                self._log('Writing ', path)
                f.write(data)
            assert os.access(path,os.R_OK)
        return

    def monitorStatus(self, param = "keys"):
        """
        Get monitor status information
        Args:
            param: query the status parameter param, if 'keys' get all status parameters.
        Returns:
            Dictionary that may contain the keys: value, value_type, unit, time, state,
            critical_limits, critical_values
        """
        return self._getRequest(self._url('monitor','status',parameter = param))

    def sendMonitorCommand(self, command):
        """
        Send monitor command to EIGER.
        Args:
            command: Command to send (up to now only "clear")
        Returns:
            Empty string
        """
        return self._putRequest(self._url("monitor","command",parameter = command), dataType = "native")
        
    def streamConfig(self,param = 'keys'):
        """
        Get stream configuration parameter
        Args:
            param: query the configuration parameter param, if 'keys' get all configuration parameters.
        Returns:
            If param is 'keys' return list of all parameters, else return dictionary
            that may contain the keys: value, min, max, allowed_values, unit, value_type and access_mode
        """
        return self._getRequest(self._url('stream','config',parameter = param))
        

    def setStreamConfig(self,param,value):
        """
        Set stream configuration parameter param.
        Args:
            param: parameter
            value: value to set
        Returns:
            List of changed parameters.
        """
        return self._putRequest(self._url('stream','config',parameter = param), dataType = 'native', data = value)
        
    def streamStatus(self, param):
        """Get stream status information
        Args:
            param: query the status parameter param, if 'keys' get all status parameters.
        Returns:
            Dictionary that may contain the keys: value, value_type, unit, time, state,
            critical_limits, critical_values
        """
        return self._getRequest(self._url('stream','status',parameter = param))
        

    def sendStreamInitialize(self):
        return self._putRequest(self._url("stream","command","initialize"), dataType = "native")

    #
    #
    #                Private Methods
    #
    #

    def _log(self,*args):
        if self._verbose:
            log.log(2,' '.join([ str(elem) for elem in args ]))

    def _url(self,module,task,parameter = None):
        url = "/{0}{1}/api/{2}/{3}/".format(self._urlPrefix,module,self._version,task)
        if not parameter is None:
            url += '{0}'.format(parameter)
        return url

    def _getRequest(self,url,dataType = 'native', fileId = None):
        if dataType is None:
            dataType = 'native'
        if dataType == 'native':
            mimeType = 'application/json; charset=utf-8'
        elif dataType == 'tif':
            mimeType = 'application/tiff'
        elif dataType == 'hdf5':
            mimeType = 'application/hdf5'
        return self._request(url,'GET',mimeType, fileId = fileId)

    def _isReadingBusy(self):
        if self.read_thread is None:
            return False
        
        is_alive = self.read_thread.is_alive()

        if is_alive is False:
            cmd_name = self.read_thread.getName()
            ret_value = self.read_thread.getAnswer()
            answer = ( cmd_name, ret_value )
            self.read_value[cmd_name] = ret_value
            self.read_thread = None

        return is_alive 

    def getRequestResponse(self, name):
        if self._isReadingBusy():
            return "busy"
        else:
            if name in self.read_value:
                return self.read_value[name]
            else:
                return ".error."

    def _putRequest(self,url,dataType,data = None):
        data, mimeType = self._prepareData(data,dataType)
        return self._request(url,'PUT',mimeType, data)

    def _delRequest(self,url):
        self._request(url,'DELETE',mimeType = None)
        return None

    def _request(self, url, method, mimeType, data = None, fileId = None):
        if data is None:
            body = ''
        else:
            body = data
        headers = {}
        if method == 'GET':
            headers['Accept'] = mimeType
        elif method == 'PUT':
            headers['Content-type'] = mimeType
        if not self._user is None:
            headers["Authorization"] = "Basic {0}".format(self._user)

        self._log('sending request to {0}'.format(url))
        numberOfTries = 0
        response = None

        try:
            while response is None:
                try:
                    self._connection.request(method,url, body = data, headers = headers)
                    response = self._connection.getresponse()
                except Exception as e:
                    numberOfTries += 1
                    if numberOfTries == 50:
                        self._log("Terminate after {0} tries\n".format(numberOfTries))
                        raise e
                    self._log("Failed to connect to host. Retrying\n")
                    self._connection = http.HTTPConnection(self._host,self._port, timeout = self._connectionTimeout)
                    continue
    
            status = response.status
            reason = response.reason
            self._log("response arrived status:%s / reason. %s" % (status,reason))
            if fileId is None:
                data = response.read()
            else:
                bufferSize = 8*1024
                while True:
                    data = response.read(bufferSize)
                    if len(data) > 0:
                        fileId.write(data)
                    else:
                        break
        except:
            self._log('request error')
        finally:
            pass
            
        mimeType = response.getheader('content-type','text/plain')
        self._log('Return status: ', status, reason)
        if not response.status in range(200,300):
            raise RuntimeError((reason,data))

        if 'json' in mimeType:
            if self._serializer is None:    
                return json.loads(data)
            else:
                return self._serializer.loads(data)
        else:
            return data

    def _parseResponse(self, response):
        status = response.status
        reason = response.reason
        self._log("parsing response status:%s / reason. %s" % (status,reason))

        data = response.read()
            
        mimeType = response.getheader('content-type','text/plain')
        
        if not response.status in range(200,300):
            raise RuntimeError((reason,data))

        if 'json' in mimeType:
            if self._serializer is None:    
                return json.loads(data)
            else:
                return self._serializer.loads(data)
        else:
            return data

    def isfile(self,obj):
        if sys.version_info[0] >= 3:
            return isinstance(obj,IOBase)
        else:
            return isinstance(obj,file)

    def _prepareData(self,data, dataType):
        if data is None:
            return '', 'text/html'

        if dataType != 'native':
            if self.isfile(data):
                data = data.read()
            if dataType is None:
                mimeType = self._guessMimeType(data)
                if not mimeType is None:
                    return data, mimeType
            elif dataType == 'tif':
                return data, 'application/tiff'

        mimeType = 'application/json; charset=utf-8'

        if self._serializer is None:
            return json.dumps({'value':data}), mimeType

        return self._serializer.dumps({"value":data}), mimeType

    def _guessMimeType(self,data):
        if type(data) == str:
            if data.startswith('\x49\x49\x2A\x00') or data.startswith('\x4D\x4D\x00\x2A'):
                self._log('Determined mimetype: tiff')
                return 'application/tiff'
            if data.startswith('\x89\x48\x44\x46\x0d\x0a\x1a\x0a'):
                self._log('Determined mimetype: hdf5')
                return 'application/hdf5'
        return None






import h5py
import sys
import os

class h5crawler_error(Exception):
    pass

class h5crawler(object):
    def __init__(self, fileobj):
        if isinstance(fileobj,str):
            self.hfile = h5py.File(fileobj)
        elif isinstance(fileobj, h5py.File):
            self.hfile = fileobj
        else:
            self.hfile = None

    # search elements with key=value. returns full_path
    def search_key(self, search_key, obj=None,kyroot=None):
        self.search_wrapper(search_func=self._search_key_function, pars={'search_key':search_key})
        return self.result

    def _search_key_function(self,pars,child,full_key,key):
        search_key = pars['search_key']
        if key == search_key:
            return full_key
        return False

    # search datasets with 2d shape
    def search_2d(self):
        self.search_wrapper(search_func=self._search_2d_function, pars=None)
        return self.result

    def _search_2d_function(self,pars,child,full_key,key):
        if isinstance(child,h5py.Dataset):
           if len(child.shape) >= 2:
              return [full_key,child.shape]

    def search_wrapper(self, search_func, pars, obj=None,kyroot=None):

        if obj is None:
            obj = self.hfile
            self.result = []
            kyroot='/'

        for ky in obj.keys():
            cky = os.path.join(kyroot,ky)
            child = self.hfile[cky]

            r = search_func(pars,child,full_key=cky,key=ky)

            if r:
                self.result.append(r)

            if not isinstance(child,h5py.Dataset):
                self.search_wrapper(search_func,pars,obj=child,kyroot=cky)

        return self.result


cr = h5crawler(sys.argv[1])

result = cr.search_2d()
#result = cr.search_key('data')
for r in result:
    print(r)





from pyspec.css_logger import log
from pyspec.client import SpecConnection
#from pyspec.client.SpecConnection import SpecConnection
import pyspec.client
import time

def test_splot():
    ldt = pyspec.client.variable(spec, 'LDT')
    print(ldt.get())
    
def test_cmd():
    cmdno = 0
    cmdno += 1
    bla = spec.run_cmd('testcnt 1')
    time.sleep(0.1)
    bla = spec.run_cmd('testcnt 2')
    time.sleep(0.1)

def test_channels():
    chan = spec.get_channel('var/MYVAR')
    print(chan.read())
    chan.write(45)
    print(chan.read())

def test_motors():
    mot = spec.get_motor('chi')
    print("Position: %s" % mot.position)
    print("Velocity: %s" % mot.read("slew_rate"))
    mot.mv(1)

def test_variables():
    pass

def test_commands():
    pass

import sys

log.start()
spec = SpecConnection(sys.argv[1])

#test_splot()
#test_channels()
#test_motors()
test_cmd()
time.sleep(1)



#
#  @(#)__init__.py	6.1  05/11/20 CSS
#  "pyspec" Release 6
#

import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from graphics_rc import g_rc

def qt_imported():
    return g_rc.qt_imported

def qt_variant():
    return g_rc.qt_variant

def qt_version():
    return g_rc.qt_version

def graph_variant():
    return g_rc.graph_variant

def mpl_imported():
    return g_rc.mpl_imported

def mpl_available():
    return g_rc.mpl_available

def mpl_version():
    return g_rc.mpl_version

def mpl_version_no():
    return g_rc.mpl_version_no

def qwt_imported():
    return g_rc.qwt_imported 




#  %W%  %G% CSS
#  "pyspec" Release %R%
#

import asyncore
import threading
import weakref
import time

try:
    import gevent
    gevent_ok = True
except:
    gevent_ok = False

THREAD, GEVENT = 0,1

from pyspec.css_logger import log
from pyspec.utils import async_loop

class spec_updater(object):
    default_update_time = 5 # millisecs

    def __init__(self, method=THREAD, update_func=None, update_time=None): 

        self.started = False
        self.thread = None
        self.greenlet = None

        if method == THREAD:
            self._method = THREAD
        elif method == GEVENT:
            if not gevent_ok:
                raise Exception("spec_updater gevent selected "
                                "but gevent module cannot be imported")
            else:
                raise Exception("gevent method is not yet implemented")
                self._method = GEVENT
                pass
        else:
            raise Exception("wrong method for spec_updater. " 
                            "only thread and gevent supported")

        self.update_func = (update_func is not None) \
                and update_func or self._update
        self.set_update_time(update_time)

    def set_update_time(self, update_time):
        self.update_time = update_time is None and \
                              self.default_update_time or update_time
        self.update_time /= 1000.0 # in seconds
        log.log(2, " starting spec updater. update time is %s" % self.update_time)


        # minimum 10 milliseconds update cycle time
        if self.update_time < 0.005:
            log.log(2, "update time of %s secs too short. using 0.01")
            self.update_time = 0.005

    def _update(self):
        # asyncore.loop(timeout=1, count=1)
        async_loop(timeout=0.01, count=1)

    def start(self):
        self.started = True
        self.start_thread()

    def stop(self,timeout=0.3):
        self.started = False
        s0 = time.time()
        try:
            while self.thread.is_alive():
                time.sleep(0.1)
                if time.time() -s0 > timeout:
                    print("cannot stop thread. killing it")
                    self.kill()
        except ImportError:
            self.kill()
            print("i am already done")

    def kill(self):
        self.started = False
        #self.thread.kill()

    def is_running(self):
        if self.thread is not None and self.thread.is_alive():
            return True
        return False

    def start_thread(self):
        self.thread = threading.Thread(target=self._loop)
        # daemon flag sets thread to stop on main thread control-C
        self.thread.daemon = True
        self.thread.start() 

    def _loop(self):
        log.log(2, "starting thread loop - update_time is %s" % self.update_time)
        while True:
            try:
                if not self.started:
                    log.log(2, "setting thread started to False. quitting")
                    break
                self.update_func()
                time.sleep(self.update_time)
            except:
                import traceback
                print(traceback.format_exc())
                break

if __name__ == '__main__':
    def say():
        print("hello")

    updater = spec_updater(update_func=say, update_time=50)
    updater.start()

    while True:
        try:
            time.sleep(0.2)
        except KeyboardInterrupt:
            break



#  %W%  %G% CSS
#  "pyspec" Release %R%
#
#$Id: SpecVariable.py,v 1.4 2005/03/17 12:43:46 guijarro Exp $

"""SpecVariable module

This module defines the class for Spec variable objects
"""
import SpecEventsDispatcher as SpecEventsDispatcher
import SpecWaitObject as SpecWaitObject

from SpecEventsDispatcher import UPDATEVALUE, FIREEVENT

class SpecVariable(object):
    """SpecVariable class

    Thin wrapper around SpecChannel objects, to make
    variables watching, setting and getting values easier.
    """
    def __init__(self, conn, varname, callbacks={}):
        """Constructor

        Keyword arguments:
        varname -- the variable name in Spec
        specapp -- 'host:port' string representing a Spec server to connect to (defaults to None)
        timeout -- optional timeout (defaults to None)
        """

        if None in (varname, conn):
            raise SpecClientError("Bad SpecVariable initialization")

        if "/" in varname: 
            self.chan_name = str(varname)
        else:
            self.chan_name = 'var/' + str(varname)

        self._callbacks = {
          'connected': None,
          'disconnected': None,
          'update': None,
        }

        self._conn = conn

        self._conn.connect_event('connected', self.spec_connected)
        self._conn.connect_event('disconnected', self.spec_disconnected)

        print("callbacks are: %s" % str(callbacks))
        for cb_name in iter(self._callbacks.keys()):
            if callable(callbacks.get(cb_name)):
               print("adding callback for signal %s" % cb_name)
               self._callbacks[cb_name] = SpecEventsDispatcher.callableObjectRef(callbacks[cb_name])

    def spec_connected(self):
        pass
    def spec_disconnected(self):
        pass

    def is_connected(self):
        """Return whether the remote Spec version is connected or not."""
        return self._conn is not None and self._conn.is_connected()

    def get(self):
        """Return the watched variable current value."""
        if self.is_connected():
            return self._conn.read_channel(self.chan_name)
        return None

    getValue = get

    @property
    def value(self):
        return self.get()

    @value.setter
    def value(self, value):
        self.set(value)

    def set(self, value):
        """Set the watched variable value

        Arguments:
        value -- the new variable value
        """
        if self.is_connected():
            return self._conn.write_channel(self.chan_name, value)

    setValue = set

    def waitUpdate(self, waitValue = None, timeout = None):
        """Wait for the watched variable value to change

        Keyword arguments:
        waitValue -- wait for a specific variable value
        timeout -- optional timeout
        """
        if self.is_connected():
            w = SpecWaitObject.SpecWaitObject(self._conn)
            w.waitChannelUpdate(self.chan_name, waitValue = waitValue, timeout = timeout)
            return w.value

class SpecVariableA(SpecVariable):
    """SpecVariableA class - asynchronous version of SpecVariable

    Thin wrapper around SpecChannel objects, to make
    variables watching, setting and getting values easier.
    """
    def __init__(self, conn, varname, dispatchMode = UPDATEVALUE, callbacks={}):
        """Constructor

        Keyword arguments:
        varname -- name of the variable to monitor (defaults to None)
        specapp -- 'host:port' string representing a Spec server to connect to (defaults to None)
        """
        super(SpecVariableA,self).__init__(conn, varname, callbacks)

        self.dispatchMode = dispatchMode

        if self._conn.is_connected():
            self.spec_connected()

    def refresh(self):
        self._conn.update()

    def spec_connected(self):
        #
        # register channel
        #
        self._conn.registerChannel(self.chan_name, self._update, dispatchMode = self.dispatchMode)

        try:
            if self._callbacks.get("connected"):
                cb = self._callbacks["connected"]()
                if cb is not None:
                    cb()
        finally:
            self.connected()

    def connected(self):
        """Callback triggered by a 'connected' event from Spec

        To be extended by derivated classes.
        """
        pass

    def spec_disconnected(self):
        try:
            if self._callbacks.get("disconnected"):
                cb = self._callbacks["disconnected"]()
                if cb is not None:
                    cb()
        finally:
            self.disconnected()

    def disconnected(self):
        """Callback triggered by a 'disconnected' event from Spec

        To be extended by derivated classes.
        """
        pass

    def _update(self, value):
        print("new value is %s" % value)
        print("toto")
        print("see if update in _callbacks: %s " % str(self._callbacks))

        try:
            if self._callbacks.get("update"):
                cb = self._callbacks["update"]()
                if cb is not None:
                    cb(value)
        finally:
            self.update(value)

    def update(self, value):
        """Callback triggered by a variable update

        Extend it to do something useful.
        """
        pass




#  @(#)SpecArray.py	3.4  05/11/20 CSS
#  "pyspec" Release 3
#

from pyspec.css_logger import log
from pyspec.utils import is_python3

import numpy

from numpy.version import version
npvers = tuple(map(int, version.split('.')))

(ARRAY_DOUBLE, ARRAY_FLOAT, ARRAY_LONG, ARRAY_ULONG, ARRAY_SHORT, \
 ARRAY_USHORT, ARRAY_CHAR, ARRAY_UCHAR, \
 ARRAY_STRING, ARRAY_NUMERIC) = (5,6,7,8,9,10,11,12,13,14)

(ARRAY_MIN, ARRAY_MAX) = (ARRAY_DOUBLE, ARRAY_STRING)

SPEC_TO_NUM = {
    ARRAY_CHAR   :  numpy.byte,
    ARRAY_UCHAR  :  numpy.ubyte,
    ARRAY_SHORT  :  numpy.short,
    ARRAY_USHORT :  numpy.ushort,
    ARRAY_LONG   :  numpy.int32,
    ARRAY_ULONG  :  numpy.uint32,
    ARRAY_FLOAT  :  numpy.float32,
    ARRAY_DOUBLE :  numpy.float64
}

NUM_TO_SPEC = {
    numpy.ubyte : ARRAY_CHAR,
    numpy.uint : ARRAY_ULONG,
    numpy.uint16 : ARRAY_USHORT,
    numpy.uint32 : ARRAY_ULONG,
    numpy.uint8 : ARRAY_CHAR,
    numpy.ushort : ARRAY_USHORT,
    numpy.short : ARRAY_SHORT,
    numpy.int32 : ARRAY_LONG,
    numpy.int8 : ARRAY_CHAR,
    numpy.float32 : ARRAY_FLOAT,
    numpy.float64 : ARRAY_DOUBLE
}

if npvers >= (1,24):
    NUM_TO_SPEC[float] = ARRAY_FLOAT
else:
    NUM_TO_SPEC[numpy.float] = ARRAY_FLOAT

class SpecArrayError(Exception):
    pass

def isArrayType(datatype):
    is_int = isinstance(datatype,int)
    return is_int and datatype >= ARRAY_MIN and datatype <= ARRAY_MAX

def SpecArray(data, datatype = ARRAY_CHAR, rows = 0, cols = 0):

    if isinstance(data, SpecArrayData):
        # create a SpecArrayData from a SpecArrayData ("copy" constructor)
        return SpecArrayData(data.data, data.type, data.shape)

    if datatype == ARRAY_STRING:
        # a list of strings
        newArray = filter(None, [x != chr(0) and x or None for x in data.split(chr(0))])
        return newArray
    else:
        newArray = None

    if isinstance(data,numpy.ndarray) :
        # convert from a Num* array to a SpecArrayData instance
        # (when you send)
        if len(data.shape) > 2:
            raise SpecArrayError("Spec arrays cannot have more than 2 dimensions")

        try:
            if type(data) == numpy.ndarray:
                numtype = data.dtype.type
                datatype = NUM_TO_SPEC[numtype]
            else:
                numtype = data.typecode()
                datatype = NUM_TO_SPEC[numtype]
        except KeyError:
            data = ''
            datatype = ARRAY_CHAR
            rows = 0
            cols = 0
            log.error("Numerical type '%s' not supported" , numtype)
        else:
            if len(data.shape) == 2:
                rows, cols = data.shape
            else:
                rows, cols = 1, data.shape[0]
            data = data.tostring()

        newArray = SpecArrayData(data, datatype, (rows, cols))
    else:
        # return a new Num* array from data
        # (when you receive)
        try:
            numtype = SPEC_TO_NUM[datatype]
        except:
            raise SpecArrayError('Invalid Spec array type')
        else:
            if is_python3(): 
                newArray = numpy.frombuffer(data, dtype=numtype)
            else:
                newArray = numpy.fromstring(data, dtype=numtype)

            if rows==1:
              newArray.shape = (cols, )
            else:
              newArray.shape = (rows, cols)

    return newArray


class SpecArrayData:
    def __init__(self, data, datatype, shape):
        self.data = data
        self.type = datatype
        self.shape = shape


    def tostring(self):
        return str(self.data)
